import potentials

from ...tools import aslist

class Database(potentials.Database):
    """
    Child of potentials.Database extended for interacting with structure
    and defect reference records.
    """
    # Class imports
    from ._crystal_prototype import (crystal_prototypes, crystal_prototypes_df,
                            load_crystal_prototypes, _no_load_crystal_prototypes,
                            get_crystal_prototypes, get_crystal_prototype,
                            download_crystal_prototypes)

    from ._relaxed_crystal import (relaxed_crystals, relaxed_crystals_df,
                            load_relaxed_crystals, _no_load_relaxed_crystals,
                            get_relaxed_crystals, get_relaxed_crystal,
                            download_relaxed_crystals)

    def __init__(self, host=None, username=None, password=None, certification=None,
                 localpath=None, verbose=False, local=None, remote=None, 
                 load=False, status='active'):
        """
        Parameters
        ----------
        host : str, optional
            CDCS site to access.  Default value is 'https://potentials.nist.gov/'.
        username : str, optional 
            User name to use to access the host site.  Default value of '' will
            access the site as an anonymous visitor.
        password : str, optional
            Password associated with the given username.  Not needed for
            anonymous access.
        certification : str, optional
            File path to certification file if needed for host.
        localpath : str, optional
            Path to the local library directory to use.  If not given, will use
            the set library_directory setting.
        verbose : bool, optional
            If True, info messages will be printed during operations.  Default
            value is False.
        local : bool, optional
            Indicates if the load operations will check localpath for records.
            Default value is controlled by settings.
        remote : bool, optional
            Indicates if the load operations will download records from the
            remote database.  Default value is controlled by settings.  If a
            local copy exists, then setting this to False is considerably
            faster.
        load : bool, str or list, optional
            If True, citations, potentials and lammps_potentials will all be
            loaded during initialization. If False (default), none will be
            loaded.  Alternatively, a str or list can be given to specify which
            of the three record types to load.
        status : str, list or None, optional
            Only potential_LAMMPS records with the given status(es) will be
            loaded.  Allowed values are 'active' (default), 'superseded', and
            'retracted'.  If None is given, then all potentials will be loaded.
        """
        # Extract extra loads options
        newoptions = ['crystal_prototype', 'relaxed_crystal']
        if load is True:
            newload = newoptions
        elif load is False:
            newload = []
        else:
            load = aslist(load)
            newload = []
            for key in newoptions:
                try:
                    newload.append(load.pop(load.index(key)))
                except:
                    pass
        
        # Call parent init
        super().__init__(host=host, username=username, password=password, certification=certification,
                 localpath=localpath, verbose=verbose, local=local, remote=remote, 
                 load=load, status=status)

        # Call new load options
        if 'crystal_prototype' in newload:
            self.load_crystal_prototypes(verbose=verbose)
        else:
            self._no_load_crystal_prototypes()

        # Call new load options
        if 'relaxed_crystal' in newload:
            self.load_relaxed_crystals(verbose=verbose)
        else:
            self._no_load_relaxed_crystals()

    def download_all(self, localpath=None, format='json', citeformat='bib',
                     indent=None, status='active', verbose=False,
                     getfiles=True):
        """
        Downloads all records from the remote to localhost.

        Parameters
        ----------
        localpath : str, optional
            Path to a local directory where the records are to be copied to.
            If not given, will check localpath value set during object
            initialization.
        format : str, optional
            The file format to save the results locally as.  Allowed values are
            'xml' and 'json'.  Default value is 'json'.
        citeformat : str, optional
            The file format to save Citation records locally as.  Allowed
            values are 'xml', 'json', and 'bib'.  Default value is 'bib'.
        indent : int, optional
            The indentation spacing size to use for the locally saved record files.
            If not given, the JSON/XML content will be compact.
        verbose : bool, optional
            If True, info messages will be printed during operations.  Default
            value is False.
        status : str, list or None, optional
            Only potential_LAMMPS records with the given status(es) will be
            downloaded.  Allowed values are 'active' (default), 'superseded', and
            'retracted'.  If None is given, then all potentials will be downloaded.
        getfiles : bool, optional
            If True, the parameter files associated with the potential_LAMMPS
            record will also be downloaded.
        
        Raises
        ------
        ValueError
            If no localpath, no potentials, invalid format, or records in a
            different format already exist in localpath.
        """
        self.download_citations(localpath=localpath, format=citeformat,
                                indent=indent, verbose=verbose)

        self.download_potentials(localpath=localpath, format=format,
                                 indent=indent, verbose=verbose)

        self.download_lammps_potentials(localpath=localpath, format=format,
                                        indent=indent, verbose=verbose,
                                        status=status, getfiles=getfiles)

        self.download_crystal_prototypes(localpath=localpath, format=format,
                                        indent=indent, verbose=verbose)


# coding: utf-8

# http://www.numpy.org/
import numpy as np

# https://matplotlib.org/
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from matplotlib import cm

# atomman imports
from ..tools import axes_check
from .. import Box, NeighborList

class DifferentialDisplacement():
    def __init__(self, system_0, system_1, neighbors=None, cutoff=None, reference=1):
        """
        Class initializer.  Calls solve if either neighbors or cutoff are given.
        
        Parameters
        ----------
        system_0 : atomman.system
            The base/reference system to use.
        system_1 : atomman.system
            The defect/current system to use.
        neighbors : atomman.NeighborList, optional
            The neighbor list to use.  
        cutoff : float, optional
            Cutoff distance for computing a neighbor list. If reference = 0, then system_0
            will be used to generate the list. If reference = 1, then system_1 will be
            used to generate the list.
        reference : int, optional
            Indicates which of the two systems should be used for the plotting
            reference: 0 or 1. If 0, then system_0's atomic positions will be
            used for the calculation and neighbors should be for system_0.  If
            1 (default), then system_1's atomic positions will be used
            for the calculation and neighbors should be for system_1.   
        """
        
        if neighbors is not None or cutoff is not None:
            self.solve(system_0, system_1, neighbors=neighbors, cutoff=cutoff, reference=reference)
        else:
            assert system_0.natoms == system_1.natoms
            
            self.__system_0 = system_0
            self.__system_1 = system_1
            self.reference = reference
            self.__neighbors = None
            self.__ddvectors = None
            self.__arrowcenters = None
            self.__arrowuvectors = None
    
    @property
    def reference(self):
        """int : Indicates which system (0 or 1) is used as the reference."""
        return self.__reference

    @reference.setter
    def reference(self, value):
        assert value == 0 or value == 1, 'reference must be 0 or 1'
        self.__reference = value    
    
    @property
    def system0(self):
        """atomman.System : The defect-free base system."""
        return self.__system0
    
    @property
    def system1(self):
        """atomman.System : The defect containing system."""
        return self.__system1
    
    @property
    def neighbors(self):
        """atomman.NeighborList : The list of neighbors identified for the reference system."""
        return self.__neighbors
    
    @property
    def ddvectors(self):
        """numpy.array or None : The computed differential displacement vectors."""
        return self.__ddvectors
    
    @property
    def arrowcenters(self):
        """numpy.array or None : The identified center positions for the ddvectors."""
        return self.__arrowcenters
    
    @property
    def arrowuvectors(self):
        """numpy.array or None : The unit vectors between all pairs of atoms for which the ddvectors have been computed."""
        return self.__arrowuvectors
    
    def solve(self, system0=None, system1=None, neighbors=None, cutoff=None, reference=None):
        """
        Solves the differential displacement vectors.
        
        Parameters
        ----------
        system0 : atomman.system
            The base/reference system to use.
        system1 : atomman.system
            The defect/current system to use.
        neighbors : atomman.NeighborList, optional
            The neighbor list to use.  
        cutoff : float, optional
            Cutoff distance for computing a neighbor list. If reference = 0, then system_0
            will be used to generate the list. If reference = 1, then system_1 will be
            used to generate the list.
        reference : int, optional
            Indicates which of the two systems should be used for the plotting reference: 0 or 1.
            If 0 (default), then system0's atomic positions will be used for the calculation and
            neighbors should be for system0.  If 1, then system1's atomic positions will be used
            for the calculation and neighbors should be for system1.   
        """
        # Handle parameters
        if system0 is not None:
            self.__system0 = system0
        else:
            system0 = self.system0
        if system1 is not None:
            self.__system1 = system1
        else:
            system1 = self.system1
        assert system0.natoms == system1.natoms
        
        if reference is None:
            reference = self.reference
        else:
            self.reference = reference
        if reference == 0:
            refsystem = system0
        else:
            refsystem = system1
        
        if neighbors is None:
            if cutoff is not None:
                self.__neighbors = neighbors = refsystem.neighborlist(cutoff=cutoff)
            else:
                if self.neighbors is not None:
                    neighbors = self.neighbors
                else:
                    raise ValueError('Either neighbors or cutoff must be given')
        else:
            self.__neighbors = neighbors
        
        all_ddvectors = []
        all_arrowcenters = []
        all_arrowuvectors = []
        
        # Loop over all atoms i in ref system
        for i in np.arange(refsystem.natoms):
            neighs = neighbors[i]
            if len(neighs) == 0:
                continue
            
            # Compute distance vectors between atom i and its neighbors for both systems
            dvectors0 = system0.dvect(int(i), neighs)
            dvectors1 = system1.dvect(int(i), neighs)
            if dvectors0.shape == (3,):
                dvectors0.shape = (1,3)
                dvectors1.shape = (1,3)

            # Compute differential displacement vectors
            ddvectors = dvectors1 - dvectors0

            # Compite center points and direction vectors
            if reference == 0:
                arrowcenters = system0.atoms.pos[i] + dvectors0 / 2
                arrowuvectors = dvectors0 / np.linalg.norm(dvectors0, axis=1)[:,np.newaxis]    
            else:    
                arrowcenters = system1.atoms.pos[i] + dvectors1 / 2
                arrowuvectors = dvectors1 / np.linalg.norm(dvectors1, axis=1)[:,np.newaxis]
                
            # Append calculation values to associated lists            
            all_ddvectors.append(ddvectors)
            all_arrowcenters.append(arrowcenters)
            all_arrowuvectors.append(arrowuvectors)
        
        # Save computed values to object properties 
        self.__ddvectors = np.concatenate(all_ddvectors)
        self.__arrowcenters = np.concatenate(all_arrowcenters)
        self.__arrowuvectors = np.concatenate(all_arrowuvectors)
        
    def plot(self, component, ddmax, plotxaxis='x', plotyaxis='y',
             xlim=None, ylim=None, zlim=None,
             arrowscale=1, arrowwidth=0.005,  use0z=False,
             atomcolor=None, atomcmap=None, atomsize=0.5, figsize=10,
             matplotlib_axes=None):

        r"""
        Creates a matplotlib figure of a differential displacement map.  Atom
        positions are represented as circles, while the selected components of the
        differential displacement vectors are plotted as arrows.
        
        Parameters
        ----------
        component : str or array-like object
            Indicates the component(s) of the differential displacement to plot.
            Values of 'x', 'y', or 'z' will plot the component along that
            Cartesian direction.  A value of 'projection' will plot the
            differential displacement vectors as projected onto the plotting
            plane, thereby showing the two components perpendicular to the line
            direction.  If a 3D vector is given, then the component parallel to
            that direction will be used.
        ddmax : float or None
            The maximum differential displacement value allowed. Values will be
            kept between +-ddmax by wrapping values with larger absolute values
            around by adding/subtracting 2*ddmax. Typically, this is set to be
            \|b\|/2, but can be defect-specific. For instance, fcc a/2<110>
            dislocations and basal hcp dislocations are typically plotted with
            ddmax=\|b\|/4.  If set to None, then no wrapping is done.
        plotxaxis : str or array-like object, optional
            Indicates the Cartesian direction associated with the system's atomic
            coordinates to align with the plotting x-axis.  Values are either 3D
            unit vectors, or strings 'x', 'y', or 'z' for the Cartesian axes
            directions.  plotxaxis and plotyaxis must be orthogonal.  Default value
            is 'x' = [1, 0, 0].
        plotyaxis : str or array-like object, optional
            Indicates the Cartesian direction associated with the system's atomic
            coordinates to align with the plotting y-axis.  Values are either 3D
            unit vectors, or strings 'x', 'y', or 'z' for the Cartesian axes
            directions.  plotxaxis and plotyaxis must be orthogonal.  Default value
            is 'y' = [0, 1, 0].
        xlim : tuple, optional
            The minimum and maximum coordinates along the plotting x-axis to
            include in the fit.  Values are taken in the specified length_unit.
            If not given, then the limits are set based on min and max atomic
            coordinates along the plotting axis.
        ylim : tuple, optional
            The minimum and maximum coordinates along the plotting y-axis to
            include in the fit.  Values are taken in the specified length_unit.
            If not given, then the limits are set based on min and max atomic
            coordinates along the plotting axis.
        zlim : tuple, optional
            The minimum and maximum coordinates normal to the plotting axes
            (i.e. plotxaxis X plotyaxis) to include in the fit.  Values are taken
            in the specified length_unit.  The optimum zlim should encompass only
            a single periodic slice.  If not given, then the limits are set
            based on min and max atomic coordinates along the axis.
        arrowscale : float, optional
            Scaling factor for the magnitude of the differential displacement
            arrows.  Default value is 1: no scaling, vectors are in units of length.
            For major components, this is often set such that the max differential
            displacement compoent after wrapping (see ddmax) is scaled to the
            distance between the atom pairs in the plot.  For minor components, this
            is often set to a large value simply to make the components visible.
        arrowwidth : float, optional
            Scaling factor to use for the width of the plotted arrows. Default value is
            0.005 = 1/200.
        use0z : bool, optional
            If False (default), the z coordinates from the reference system will be
            used for zlim and atomcmap colors. If True, the z coordinates will be
            used from system0 even if system1 is the reference system.
        atomcolor : str or list, optional
            Matplotlib color name(s) to use to display the atoms.  If str, that
            color will be assigned to all atypes.  If list, must give a color value
            or None for each atype.  Default value (None) will use cmap instead.
            Note: atomcolor and atomcmap can be used together as long as exactly
            one color or cmap is given for each unique atype.
        atomcmap : str or list, optional
            Matplotlib colormap name(s) to use to display the atoms.  Atoms will
            be colored based on their initial positions and scaled using zlim. If
            str, that cmap will be assigned to all atypes.  If list, must give a 
            cmap value or None for each atype.  Default value (None) will use 'hsv'
            cmap.  Note: atomcolor and atomcmap can be used together as long as
            exactly one color or cmap is given for each unique atype.
        atomsize : float, optional
            The circle radius size to use for the plotted atom positions in units of
            length.  Default value is 0.5.
        figsize : float or tuple, optional
            Specifies the size of the figure to create in inches.  If a single value
            is given, it will be used for the figure's width, and the height will be
            scaled based on the xlim and ylim values.  Alternatively, both the width
            and height can be set by passing a tuple of two values, but the plot will
            not be guaranteed to be "regular" with respect to length dimensions.
        matplotlib_axes : matplotlib.Axes.axes, optional
            An existing matplotlib axes object. If given, the differential displacement
            plot will be added to the specified axes of an existing figure.  This
            allows for subplots to be constructed.  Note that figsize will be ignored
            as the figure would have to be created beforehand and no automatic
            optimum scaling of the figure's dimensions will occur.
            
        Returns
        -------
        matplotlib.Figure
            The generated figure.  Not returned if matplotlib_axes is given.
        """

        ###################### Parameter and plot setup ########################

        # Interpret plot axis values
        plotxaxis = self.__plotaxisoptions(plotxaxis)
        plotyaxis = self.__plotaxisoptions(plotyaxis)

        # Build transformation matrix, T, from plot axes.
        T = axes_check([plotxaxis, plotyaxis, np.cross(plotxaxis, plotyaxis)])

        # Extract positions and transform using T        
        if self.reference == 0:        
            atompos = np.inner(self.system0.atoms.pos, T)
            refsystem = self.system0
        else:
            atompos = np.inner(self.system1.atoms.pos, T)
            refsystem = self.system1
            if use0z:
                pos0 = np.inner(self.system0.atoms.pos, T)
                atompos[:, 2] = pos0[:, 2]

        # Set default plot limits
        if xlim is None:
            xlim = (atompos[:, 0].min(), atompos[:, 0].max())
        if ylim is None:
            ylim = (atompos[:, 1].min(), atompos[:, 1].max())
        if zlim is None:
            zlim = (atompos[:, 2].min(), atompos[:, 2].max()) 

        # Define box for identifying only points inside
        plotbox = Box(xlo=xlim[0]-5, xhi=xlim[1]+5,
                      ylo=ylim[0]-5, yhi=ylim[1]+5,
                      zlo=zlim[0], zhi=zlim[1])
            
        # Set plot height if needed
        if isinstance(figsize, (int, float)):
            dx = xlim[1] - xlim[0]
            dy = ylim[1] - ylim[0]
            figsize = (figsize, figsize * dy / dx)

        # Initial plot setup and parameters
        if matplotlib_axes is None:
            fig = plt.figure(figsize=figsize, dpi=72)
            ax1 = fig.add_subplot(111)
        else:
            ax1 = matplotlib_axes
        ax1.axis([xlim[0], xlim[1], ylim[0], ylim[1]])
        
        # Handle atomcolor and atomcmap values
        atomcolor, atomcmap = self.__atomcoloroptions(atomcolor, atomcmap)

        ######################## Add atom circles to plot ##############################
        
        # Loop over all atoms i in plotting box
        for i in np.arange(refsystem.natoms)[plotbox.inside(atompos)]:
            atype = refsystem.atoms.atype[i]
            atype_index = refsystem.atypes.index(atype)

            # Plot a circle for atom i
            if atomcmap[atype_index] is not None:
                color = atomcmap[atype_index]((atompos[i, 2] - zlim[0]) / (zlim[1] - zlim[0]))
            elif atomcolor[atype_index] is not None:
                color = atomcolor[atype_index]
            else:
                color = None
            if color is not None:
                ax1.add_patch(mpatches.Circle(atompos[i, :2], atomsize, fc=color, ec='k'))

        ######################## Arrow setup ##############################
        
        # Build arrows based on component
        arrowlengths, arrowcenters = self.__buildarrows(T, plotbox, component, ddmax)
        
        # Scale arrows
        arrowlengths = arrowscale * arrowlengths
        
        # Compute arrow widths based on lengths
        arrowwidths = arrowwidth * (arrowlengths[:,0]**2 + arrowlengths[:,1]**2)**0.5

        # Plot the arrows
        for center, length, width in zip(arrowcenters, arrowlengths, arrowwidths):
            if width > 1e-7:
                ax1.quiver(center[0], center[1], length[0], length[1],
                           pivot='middle', angles='xy', scale_units='xy',
                           scale=1, width=width, minshaft=2)
        
        if matplotlib_axes is None:
            return fig
    
    def __plotaxisoptions(self, plotaxis):
        """Internal method for handling plotxaxis and plotyaxis values"""
        
        # Give numeric values for str plot axis terms
        if plotaxis == 'x':
            plotaxis = [1.0, 0.0, 0.0]
        elif plotaxis == 'y':
            plotaxis = [0.0, 1.0, 0.0]
        elif plotaxis == 'z':
            plotaxis = [0.0, 0.0, 1.0]
        
        # Convert to numpy array
        return np.asarray(plotaxis, dtype=float)
    
    def __atomcoloroptions(self, atomcolor, atomcmap):
        """Internal method for handling atomcolor and atomcmap options"""
        
        # Identify number of atom types in the reference system
        if self.reference == 0:
            natypes = self.system0.natypes
        else:
            natypes = self.system1.natypes
        
        # Set default atomcmap
        if atomcolor is None and atomcmap is None:
            atomcmap = 'hsv'

        # Transform single color/cmap values to lists
        if isinstance(atomcmap, str):
            if atomcolor is not None:
                raise TypeError('atomcmap and atomcolor cannot be str if both are given')
            atomcmap = [atomcmap for i in range(natypes)]
            atomcolor = [None for i in range(natypes)]
        elif isinstance(atomcolor, str):
            if atomcmap is not None:
                raise TypeError('atomcmap and atomcolor cannot be str if both are given')
            atomcolor = [atomcolor for i in range(natypes)]
            atomcmap = [None for i in range(natypes)]
        else:
            atomcolor = list(atomcolor)
            atomcmap = list(atomcmap)

        # Check atomcolor, atomcmap list compatibility
        if len(atomcmap) != natypes:
            raise ValueError('Invalid number of atomcmap values')
        if len(atomcolor) != natypes:
            raise ValueError('Invalid number of atomcolor values')
        for ic in range(len(atomcmap)):
            if atomcmap[ic] is not None:
                if atomcolor[ic] is not None:
                    raise ValueError('atomcmap and atomcolor cannot both be given for same atype')
                atomcmap[ic] = cm.get_cmap(atomcmap[ic])
        
        return atomcolor, atomcmap
    
    def __buildarrows(self, T, plotbox, component, ddmax):
        """Internal method for building the parameters for plotting the arrows"""
        
        # Manage component
        if isinstance(component, str):
            if component == 'x':
                component = np.array([1.0, 0.0, 0.0])
            elif component == 'y':
                component = np.array([0.0, 1.0, 0.0])
            elif component == 'z':
                component = np.array([0.0, 0.0, 1.0])
            elif component != 'projection':
                raise ValueError('Invalid component style: must be x, y, z, projection, or numpy array')
        else:
            component = np.asarray(component, dtype=float)
            assert component.shape == (3,), 'Invalid numeric component: must be a 3D vector'
            component = component / np.linalg.norm(component)
            
        # Transform arrow-related vectors
        ddvectors = np.inner(self.ddvectors, T)
        arrowcenters = np.inner(self.arrowcenters, T)
        arrowuvectors = np.inner(self.arrowuvectors, T)
                
        # Identify only vectors in ploting box
        inbounds = plotbox.inside(arrowcenters)
        ddvectors = ddvectors[inbounds]
        arrowcenters = arrowcenters[inbounds]
        arrowuvectors = arrowuvectors[inbounds]

        # Build arrows for the xy component option
        if isinstance(component, str) and component == 'projection':
            
            # Arrows are in-plane vector components
            ddcomponents = (ddvectors[:,0]**2 + ddvectors[:,1]**2)**0.5
            arrowuvectors = ddvectors[:, :2] / ddcomponents[:,np.newaxis]
            
            # Scheme for direction uniqueness (not sure what other projects use?)
            arrowuvectors[ddvectors[:, 2] > 0] *= -1
            
            # Normalize ddcomponents to be between +-ddmax
            if ddmax is not None and ddmax > 0:
                while True:
                    mask = ddcomponents > ddmax
                    if np.sum(mask) == 0:
                        break
                    ddcomponents[mask] -= 2 * ddmax
                while True:
                    mask = ddcomponents < -ddmax
                    if np.sum(mask) == 0:
                        break
                    ddcomponents[mask] += 2 * ddmax

            # Arrows have magnitude of ddcomponent
            arrowlengths = arrowuvectors * ddcomponents[:,np.newaxis]
        
        # Build arrows for vector components
        else:
            component = T.dot(component)
            ddcomponents = ddvectors.dot(component)

            # Normalize ddcomponents to be between +-ddmax
            if ddmax is not None and ddmax > 0:
                while True:
                    mask = ddcomponents > ddmax
                    if np.sum(mask) == 0:
                        break
                    ddcomponents[mask] -= 2 * ddmax
                while True:
                    mask = ddcomponents < -ddmax
                    if np.sum(mask) == 0:
                        break
                    ddcomponents[mask] += 2 * ddmax

            # Arrows have magnitude of ddcomponent and direction of uvectors
            arrowlengths = arrowuvectors * ddcomponents[:,np.newaxis]
        
        return arrowlengths, arrowcenters



# coding: utf-8

# http://www.numpy.org/
import numpy as np

# atomman imports
from ... import Atoms, Box, System

# https://atztogo.github.io/phonopy/
try:
    import phonopy
    has_phonopy = True
except:
    has_phonopy = False

def load(phonopyatoms, symbols=None, prop={}):
    """
    Convert a phonopy.structure.atoms.PhonopyAtoms into an atomman.System.
    
    Parameters
    ----------
    phonopyatoms : phonopy.structure.atoms.PhonopyAtoms
        A phonopy representation of a collection of atoms, based on ase.Atoms.
    symbols : tuple, optional
        Allows the list of element symbols to be assigned during loading.
        Useful if the symbols for the model differ from the standard element
        tags.
    prop : dict, optional
        Dictionary containing any extra per-atom properties to include.
        
    Returns
    -------
    system : atomman.System
        A atomman representation of a system.
    """
    
    assert has_phonopy, 'phonopy not imported'
    
    # Get box/cell information
    box = Box(vects = phonopyatoms.get_cell())
    pbc = (True, True, True)
    
    # Get element information
    all_elements = np.array(phonopyatoms.get_chemical_symbols())
    elements, atype = np.unique(all_elements, return_inverse = True)
    if symbols is None:
        symbols = elements

    # Get atomic information
    prop['atype'] = atype + 1
    prop['pos'] = phonopyatoms.get_positions()
    atoms = Atoms(prop=prop)
    
    # Build system
    system = System(atoms=atoms, box=box, pbc=pbc, symbols=symbols)
    
    return system


# coding: utf-8

# http://www.numpy.org/
import numpy as np

__all__ = ['iscubic', 'ishexagonal', 'istetragonal', 'isrhombohedral',
           'isorthorhombic', 'ismonoclinic', 'istriclinic', 'identifyfamily']

def iscubic(box, rtol=1e-05, atol=1e-08):
    """
    Tests if a box is consistent with a standard cubic cell:
    a = b = c
    alpha = beta = gamma = 90
    
    Parameters
    ----------
    box : atomman.Box
        The box object to test.
    rtol : float, optional
        Relative tolerance for testing box parameters. Default value is 1e-5.
    atol : float, optional
        Absolute tolerance for testing box parameters. Default value is 1e-8.
        
    Returns
    -------
    bool
        True if box is a standard cubic cell, False otherwise.
    """
    return (np.isclose(box.a, box.b, atol=atol, rtol=rtol)
            and np.isclose(box.a, box.c, atol=atol, rtol=rtol)
            and np.isclose(box.alpha, 90.0, atol=atol, rtol=rtol)
            and np.isclose(box.beta, 90.0, atol=atol, rtol=rtol)
            and np.isclose(box.gamma, 90.0, atol=atol, rtol=rtol))

def ishexagonal(box, rtol=1e-05, atol=1e-08):
    """
    Tests if a box is consistent with a standard hexagonal cell:
    a = b != c
    alpha = beta = 90
    gamma = 120
    
    Parameters
    ----------
    box : atomman.Box
        The box object to test.
    rtol : float, optional
        Relative tolerance for testing box parameters. Default value is 1e-5.
    atol : float, optional
        Absolute tolerance for testing box parameters. Default value is 1e-8.
        
    Returns
    -------
    bool
        True if box is a standard hexagonal cell, False otherwise.
    """
    return (np.isclose(box.a, box.b, atol=atol, rtol=rtol)
            and np.isclose(box.alpha, 90.0, atol=atol, rtol=rtol)
            and np.isclose(box.beta, 90.0, atol=atol, rtol=rtol)
            and np.isclose(box.gamma, 120.0, atol=atol, rtol=rtol))

def istetragonal(box, rtol=1e-05, atol=1e-08):
    """
    Tests if a box is consistent with a standard tetragonal cell:
    a = b != c
    alpha = beta = gamma = 90
    
    Parameters
    ----------
    box : atomman.Box
        The box object to test.
    rtol : float, optional
        Relative tolerance for testing box parameters. Default value is 1e-5.
    atol : float, optional
        Absolute tolerance for testing box parameters. Default value is 1e-8.
        
    Returns
    -------
    bool
        True if box is a standard tetragonal cell, False otherwise.
    """
    return (np.isclose(box.a, box.b, atol=atol, rtol=rtol)
            and not np.isclose(box.a, box.c, atol=atol, rtol=rtol)
            and np.isclose(box.alpha, 90.0, atol=atol, rtol=rtol)
            and np.isclose(box.beta, 90.0, atol=atol, rtol=rtol)
            and np.isclose(box.gamma, 90.0, atol=atol, rtol=rtol))

def isrhombohedral(box, rtol=1e-05, atol=1e-08):
    """
    Tests if a box is consistent with a standard rhombohedral cell:
    a = b = c
    alpha = beta = gamma != 90
    
    Parameters
    ----------
    box : atomman.Box
        The box object to test.
    rtol : float, optional
        Relative tolerance for testing box parameters. Default value is 1e-5.
    atol : float, optional
        Absolute tolerance for testing box parameters. Default value is 1e-8.
        
    Returns
    -------
    bool
        True if box is a standard rhombohedral cell, False otherwise.
    """
    return (np.isclose(box.a, box.b, atol=atol, rtol=rtol)
            and np.isclose(box.a, box.c, atol=atol, rtol=rtol)
            and np.isclose(box.alpha, box.beta, atol=atol, rtol=rtol)
            and np.isclose(box.alpha, box.gamma, atol=atol, rtol=rtol)
            and not np.isclose(box.alpha, 90.0, atol=atol, rtol=rtol))

def isorthorhombic(box, rtol=1e-05, atol=1e-08):
    """
    Tests if a box is consistent with a standard orthorhombic cell:
    a != b != c
    alpha = beta = gamma = 90
    
    Parameters
    ----------
    box : atomman.Box
        The box object to test.
    rtol : float, optional
        Relative tolerance for testing box parameters. Default value is 1e-5.
    atol : float, optional
        Absolute tolerance for testing box parameters. Default value is 1e-8.
        
    Returns
    -------
    bool
        True if box is a standard orthorhombic cell, False otherwise.
    """
    return (not np.isclose(box.a, box.b, atol=atol, rtol=rtol)
            and not np.isclose(box.a, box.c, atol=atol, rtol=rtol)
            and np.isclose(box.alpha, 90.0, atol=atol, rtol=rtol)
            and np.isclose(box.beta, 90.0, atol=atol, rtol=rtol)
            and np.isclose(box.gamma, 90.0, atol=atol, rtol=rtol))

def ismonoclinic(box, rtol=1e-05, atol=1e-08):
    """
    Tests if a box is consistent with a standard monoclinic cell:
    a != b != c
    alpha = gamma = 90
    beta != 90
    
    Parameters
    ----------
    box : atomman.Box
        The box object to test.
    rtol : float, optional
        Relative tolerance for testing box parameters. Default value is 1e-5.
    atol : float, optional
        Absolute tolerance for testing box parameters. Default value is 1e-8.
        
    Returns
    -------
    bool
        True if box is a standard monoclinic cell, False otherwise.
    """
    return (not np.isclose(box.a, box.b, atol=atol, rtol=rtol)
            and not np.isclose(box.a, box.c, atol=atol, rtol=rtol)
            and np.isclose(box.alpha, 90.0, atol=atol, rtol=rtol)
            and not np.isclose(box.beta, 90.0, atol=atol, rtol=rtol)
            and np.isclose(box.gamma, 90.0, atol=atol, rtol=rtol))

def istriclinic(box, rtol=1e-05, atol=1e-08):
    """
    Tests if a box is consistent with a standard triclinic cell:
    a != b != c
    alpha != 90
    beta != 90
    gamma != 90
    
    Parameters
    ----------
    box : atomman.Box
        The box object to test.
    rtol : float, optional
        Relative tolerance for testing box parameters. Default value is 1e-5.
    atol : float, optional
        Absolute tolerance for testing box parameters. Default value is 1e-8.
        
    Returns
    -------
    bool
        True if box is a standard triclinic cell, False otherwise.
    """
    return (not np.isclose(box.a, box.b, atol=atol, rtol=rtol)
            and not np.isclose(box.a, box.c, atol=atol, rtol=rtol)
            and not np.isclose(box.alpha, box.beta, atol=atol, rtol=rtol)
            and not np.isclose(box.alpha, box.gamma, atol=atol, rtol=rtol))

def identifyfamily(box, rtol=1e-05, atol=1e-08):
    """
    Tests if a box is consistent with a standard representation
    of a crystal system cell.
    
    Parameters
    ----------
    box : atomman.Box
        The box object to test.
    rtol : float, optional
        Relative tolerance for testing box parameters. Default value is 1e-5.
    atol : float, optional
        Absolute tolerance for testing box parameters. Default value is 1e-8.
        
    Returns
    -------
    str or None
        'cubic', 'hexagonal', 'tetragonal', 'rhombohedral', 'orthorhombic',
        'monoclinic' or 'triclinic' if it matches any. None if no matches.
        
    Raises
    ------
    ValueError
        If box is not consistent with a standard cell.
    """
    if iscubic(box, rtol=1e-05, atol=1e-08):
        return 'cubic'
    elif ishexagonal(box, rtol=1e-05, atol=1e-08):
        return 'hexagonal'
    elif istetragonal(box, rtol=1e-05, atol=1e-08):
        return 'tetragonal'
    elif isrhombohedral(box, rtol=1e-05, atol=1e-08):
        return 'rhombohedral'
    elif isorthorhombic(box, rtol=1e-05, atol=1e-08):
        return 'orthorhombic'
    elif ismonoclinic(box, rtol=1e-05, atol=1e-08):
        return 'monoclinic'
    elif istriclinic(box, rtol=1e-05, atol=1e-08):
        return 'triclinic'
    else:
        None


# coding: utf-8

# Read version from VERSION file
try:
    from importlib import resources
except:
    from pathlib import Path
    with open(Path(Path(__file__).resolve().parent, 'VERSION')) as version_file:
        __version__ = version_file.read().strip()
else:
    __version__ = resources.read_text('atomman', 'VERSION').strip()

# potentials imports
from potentials import build_lammps_potential, Settings

# atomman imports
from . import unitconvert
from . import tools
from . import mep
from . import region
from . import lammps
from .dump import *
from .dump import __all__ as dump_all
from .core import *
from .core import __all__ as core_all
from . import library
from .library import load_lammps_potential
from .load import *
from .load import __all__ as load_all
from . import plot
from . import defect


__all__ = ['__version__'] + dump_all + core_all + load_all
__all__ += ['load_lammps_potential', 'build_lammps_potential']
__all__.sort()

# Define default working units
unitconvert.reset_units(length = 'angstrom', mass = 'amu', energy='eV', charge='e')


# coding: utf-8
# Standard Python imports
import time

# http://www.numpy.org/
import numpy as np

from scipy.interpolate import CubicSpline

from . import BasePath
from ..tools import aslist

class ISMPath(BasePath):
    """
    Class representing an energy path for use with the improved string method.
    """
    
    @property
    def default_timestep(self):
        """float : The default relaxation timestep"""
        
        # 0.05 * min(0.2, N^-1)
        return 0.05 * np.min([0.2, len(self.coord)**-1])
    
    @property
    def default_tolerance(self):
        """float : The default relaxation tolerance"""
        
        # max(N^-4, 1e-10)
        return np.max([len(self.coord)**-4, 1e-10])
    
    @property
    def unittangent(self):
        """numpy.NDArray : The tangent vectors along the path at each point."""
        
        τ = np.empty_like(self.coord)
        
        # Calculate all unit difference vectors
        diff = self.coord[1:] - self.coord[:-1]
        diff = (diff.T / np.linalg.norm(diff, axis=-1)).T
        
        # Set end points using only forward, backward
        τ[0] = diff[0]
        τ[-1] = diff[-1]
        
        # Set mid points as average of forward, backward
        τ[1:-1] = diff[:-1] + diff[1:]
        
        # Normalize all to unit vectors
        τ = (τ.T / np.linalg.norm(τ, axis=-1)).T
        
        return τ
    
    def interpolate_path(self, arccoord):
        """
        Uses cubic spline interpolation to interpolate a new path from intermediate
        arc length coordinates along the current path.
        
        Parameters
        ----------
        arccoord: array-like object
            Arc length coordinates along the current path where the new points are
            to be placed.  Values must be in the range [0, self.arccoord].
            
        Returns
        -------
        Path
            A new path with the interpolated coordinates.
        """
        α = self.arccoord
        
        if np.any(arccoord < 0) | np.any(arccoord > α[-1]):
            raise ValueError(f'arccoord values must be in range [0.0, {α[-1]}]')
            
        newcoord = CubicSpline(α, self.coord)(arccoord)
        
        return ISMPath(newcoord, self.energyfxn, gradientfxn=self.gradientfxn,
                    gradientkwargs=self.gradientkwargs)

    def step(self, timestep=None, climbindex=None):
        """
        Performs a single string relaxation step.
        
        Parameters
        ----------
        timestep : float, optional
            The size of the timestep to use.  Will use the path's
            default timestep if not given.
        climbindex : int or list, optional
            Indicates the indices of the path points to apply the climb
            algorithm to.  If None, no climbing will be performed.
        
        Returns
        -------
        newpath : Path
            A Path with coordinates evolved forward by one timestep.
        """
        
        # Set default timestep
        if timestep is None:
            timestep = self.default_timestep
        
        # Define the relaxation rate functions
        def rate(coord):
            """Regular motions are along -grad(E)"""
            grad_energy = self.grad_energy(coord)
            return - grad_energy
            
        def climbrate(coord, τ):
            """Climb motions are along -grad(E) + 2 (grad(E) dot τ) τ"""
            grad_energy = self.grad_energy(coord)
            r = - grad_energy + 2 * np.einsum('ij,ij,il->il', grad_energy, τ, τ)
            return r
        
        # Integrate path coordinates using rate
        icoord = self.integratorfxn(rate, self.coord, timestep)
        
        # Integrate the climbing coordinates using climbrate
        if climbindex is not None:
            climbindex = aslist(climbindex)
            
            # Get tangent vectors for initial path
            τ = self.unittangent
            
            # Replace intpath coords for climbindices
            icoord[climbindex] = self.integratorfxn(climbrate,
                                                    self.coord[climbindex],
                                                    timestep,
                                                    τ=τ[climbindex])
        else:
            climbindex = []
        
        # Create intpath from integrated coords
        intpath = ISMPath(icoord, self.energyfxn, self.gradientfxn,
                          self.gradientkwargs)
        
        # Divide full path into segments based on climbindices
        startindices = [0] + aslist(climbindex)
        endindices = aslist(np.asarray(climbindex)+1) + [None]
        
        # Redistribute arclength coords evenly within each segment
        α = intpath.arccoord

        newα = np.empty_like(α)
        for s, e in zip(startindices, endindices):
            
            # Adjust coords in segment to be equally spaced
            subα = α[s:e]
            newα[s:e] = np.linspace(subα[0], subα[-1], len(subα))
        
        # Interpolate new path from the adjusted arclength coords
        newpath = intpath.interpolate_path(newα)
        
        return newpath
    
    def relax(self, relaxsteps=0, climbsteps=0, timestep=None,
              tolerance=None, climbpoints=1, verbose=True):
        """
        Perform multiple relaxation and/or climb steps until either the
        maximum coordinate displacement per step drops below a tolerance or
        the maximum number of steps is reached.
        
        Parameters
        ----------
        relaxsteps : int, optional
            The maximum number of relaxation steps to perform.  Default value
            is 0: no relaxation steps.
        climbsteps : int, optional
            The maximum number of climbing steps to perform.  Default value
            is 0: no climbing steps.
        timestep : float, optional
            The size of the timestep to use.  Will use default_timestep if not
            given.
        tolerance : float, optional
            The coordinate displacement tolerance to use.  Will use
            default_tolerance if not given.
        climbpoints : int, optional
            Indicates the maximum number of points to subject the climbing to.
            Default value is 1: i.e. only one maximum is refined.
        verbose : bool, optional
            If True (default), informative statements about the relaxation are
            printed.
        """
        
        # Set default timestep
        if timestep is None:
            timestep = self.default_timestep
        
        # Set default tolerance
        if tolerance is None:
            tolerance = self.default_tolerance

        if verbose:
            print(f'timestep =  {timestep}')
            print(f'tolerance = {tolerance}')
            print(flush=True)
        
        # Set current path
        currentpath = self

        # ----------------- Relaxation steps ---------------- #

        if verbose and relaxsteps > 0:
            print('Starting relaxation steps', flush=True)
        
        # Perform the relaxation steps
        s = time.time()
        for i in range(relaxsteps):
            
            # Step forward
            newpath = currentpath.step(timestep=timestep)
            
            # Calculate max displacement from step
            d = np.linalg.norm(newpath.coord - currentpath.coord, axis=-1).max() / timestep
            
            # Update currentpath to newpath
            currentpath = newpath

            # Stop if tolerance reached
            if d < tolerance:
                break
        
        e = time.time()
        if verbose and relaxsteps > 0:
            print(f'Number of relax steps performed: {i+1}')
            print(f'Final max displacement: {d}')
            print(f'Run time: {e-s} seconds')
            print(flush=True)
        
        # --------- Identify climb coordiates(s) -------- #
        
        # Identify the points that are local energy maxima
        energy = currentpath.energy()
        maxmap = np.hstack([False, (energy[1:-1] > energy[:-2]) & (energy[1:-1] > energy[2:]), False])
        climbindex = np.arange(len(maxmap))[maxmap]
        
        if maxmap.sum() > climbpoints:
            climbindex = climbindex[:climbpoints]
        
        if verbose and climbsteps > 0:
            print(f'Starting climbing steps with {len(climbindex)} climbing points', flush=True)
        
        # ----------------- Climbing steps ---------------- #
        
        # Perform the climb steps
        s = time.time()
        for i in range(climbsteps):
            
            # Step forward
            newpath = currentpath.step(timestep=timestep, climbindex=climbindex)
            
            # Calculate max displacement
            d = np.linalg.norm(newpath.coord - currentpath.coord, axis=-1).max() / timestep
            
            # Update currentpath to newpath
            currentpath = newpath
            
            # Stop if tolerance reached
            if d < tolerance:
                break
        
        e = time.time()
        if verbose and climbsteps > 0:
            print(f'Number of climb steps performed: {i+1}')
            print(f'Final max displacement: {d}')
            print(f'Run time: {e-s} seconds')
            print(f'Max energy before climb = {energy.max()}', flush=True)
            print(f'Max energy after climb = {currentpath.energy().max()}')
            print(flush=True)
        
        return currentpath


# coding: utf-8
# Standard Python libraries
import datetime

# http://www.numpy.org/
import numpy as np

# https://pandas.pydata.org/
import pandas as pd

# https://github.com/usnistgov/DataModelDict
from DataModelDict import DataModelDict as DM

# atomman imports
from ..tools import uber_open_rmode

class Simulation():
    """
    Object for representing the LAMMPS log output for a single MS/MD run
    """

    def __init__(self, thermo=None, performance=None):
        """
        Initializes a Simulation object with performance information
        """
        self.__thermo = None
        self.__performance = None
        self.__keys = []
        if thermo is not None:
            self.thermo = thermo
        if performance is not None:
            self.performance = performance

    def keys(self):
        """List of attribute keys that have been set"""
        return tuple(self.__keys)

    def __getitem__(self, key):
        """Accesses class attributes as dict keys""" 

        if key not in self.keys():
            raise KeyError(key)

        return getattr(self, key)

    def __iter__(self):
        """Iterates over set attribute keys"""
        for key in self.keys():
            yield(key)

    @property
    def thermo(self):
        """pandas.DataFrame: The Simulation's thermo data"""
        return self.__thermo

    @property
    def performance(self):
        """pandas.DataFrame: The Simulation's performance data"""
        return self.__performance

    @thermo.setter
    def thermo(self, value):
        if isinstance(value, pd.DataFrame):
            self.__thermo = value
        else:
            self.__thermo = pd.DataFrame(value)
        if 'thermo' not in self.keys():
            self.__keys.append('thermo')
            
    @performance.setter
    def performance(self, value):
        if isinstance(value, pd.DataFrame):
            self.__performance = value
        else:
            self.__performance = pd.DataFrame(value)
        if 'performance' not in self.keys():
            self.__keys.append('performance')

class Log():
    """Object for representing a LAMMPS log output"""
    
    def __init__(self, log_info=None):
        """
        Initializes a Log object.
        
        Parameters
        ----------
        log_info : str or file-like object, optional
            The LAMMPS log content to read in.  If None (default), then the
            Log object is created but empty.
        """
        
        # Initialize simulation properties
        self.__simulations = []
        self.__lammps_version = None
        self.__lammps_date = None
        
        # Read log data if supplied
        if log_info is not None:
            self.read(log_info)
    
    def read(self, log_info, append=True):
        """
        Parses a LAMMPS screen output/log file.
        
        Parameters
        ----------
        log_info : str or file-like object
            The LAMMPS log content to read in.
        append : bool, optional
            Flag indicating if the content being read in is appended to the
            current data (True, default), or if it overwrites any existing
            saved content (False).
        """
        
        # Reset properties and values if append is False
        if append is False:
            self.__simulations = []
            self.__lammps_version = None
            self.__lammps_date = None
        
        # Strings found directly before and after run and minimize thermo data
        thermo_start_trigger = ['Memory usage per processor =',
                             'Per MPI rank memory allocation (min/avg/max) =']
        thermo_end_trigger = ['Loop time of']
        
        # Strings found directly before and after run and minimize performance data
        performance_start_trigger = ['MPI task timing breakdown']
        performance_start_trigger_old_version = ['Pair  time (%)']
        performance_end_trigger = ['Nlocal:']
        is_old_version = False

        # Handle file names, strings and open file-like objects equivalently
        with uber_open_rmode(log_info) as log_info:
            
            # Initialize parameters
            thermo_headers = []
            thermo_footers = []
            performance_headers = []
            performance_footers = []
            i = 0
            
            # For all lines in file/output
            for line in log_info:
                line = line.decode('UTF-8')
                
                # Skip blank lines
                if len(line.split()) == 0:
                    continue
                
                # Save the LAMMPS version information
                if line[:8] == 'LAMMPS (' and self.lammps_version is None:
                    self.__read_lammps_version(line)
                
                # Check for strings listed prior to run and minimize simulations
                if any([trigger in line for trigger in thermo_start_trigger]):
                    thermo_headers.append(i+1)
                
                # Check for strings listed after run and minimize simulations
                elif any([trigger in line for trigger in thermo_end_trigger]):
                    thermo_footers.append(i-1)

                # Check for strings listed prior to  performance data
                if any([trigger in line for trigger in performance_start_trigger]):
                    performance_headers.append(i+1)
                if any([trigger in line for trigger in performance_start_trigger_old_version]):
                    performance_headers.append(i)
                    is_old_version = True
                
                # Check for strings listed after performance data
                elif any([trigger in line for trigger in performance_end_trigger]):
                    performance_footers.append(i-1)
                
                i += 1
            
            # Add last line to footers for incomplete logs
            thermo_footers.append(i)
            
            # Reset file pointer
            log_info.seek(0)
            
            # Read thermo data and create Simulations
            for header, footer in zip(thermo_headers, thermo_footers):
                self.__read_thermo(log_info, header, footer)

            # Read performance data for each Simulation
            for i in range(len(performance_footers)):
                header = performance_headers[i]
                footer = performance_footers[i]
                performance = self.__read_performance(log_info, header, footer, is_old_version)
                self.simulations[i].performance = performance
            
    def __read_lammps_version(self, line):
        """
        Subfunction for reading the LAMMPS version from the log file
        """
        month = {'Jan': 1, 'Feb': 2, 'Mar': 3, 'Apr': 4, 
                 'May': 5, 'Jun': 6, 'Jul': 7, 'Aug': 8,
                 'Sep': 9, 'Oct': 10,'Nov': 11,'Dec': 12}
        self.__lammps_version = line.strip()[8:-1]
        d = self.lammps_version.split('-')[0].split()
        self.__lammps_date = datetime.date(int(d[2]), month[d[1]], int(d[0]))

    def __read_thermo(self, log_info, header, footer):
        """
        Subfunction for reading the thermo data associated with a simulation run.
        """
        # Use pandas to read all thermo data at once
        thermo = pd.read_csv(log_info, header=header,
                                nrows=footer-header,
                                delim_whitespace=True,
                                skip_blank_lines=True)      
        
        # Reset file pointer
        log_info.seek(0)

        # Append simulation results
        self.__simulations.append(Simulation(thermo=thermo))

    def __read_performance(self, log_info, header, footer, is_old_version):
        """
        Subfunction for reading the performance data associated with a simulation run.
        """

        # Use pandas to read all performance data at once        
        if not is_old_version:
            performance = pd.read_csv(log_info, header=header,
                                nrows=footer-header,
                                sep = '|',
                                skip_blank_lines=True)  
            performance = performance.drop([0])
            performance.rename(columns=lambda x: x.strip(), inplace=True)
            performance = performance.set_index('Section')
            performance = performance.replace(r'^\s*$',0.0,regex=True)
            performance = performance.astype(float) 
        
        else: 
            performance = pd.read_csv(log_info, header=header,
                                nrows=footer-header,
                                sep = '=',
                                skip_blank_lines=True)
            performance = performance.columns.to_frame().T.append(performance,ignore_index=True)
            performance.rename(columns=lambda x: x.strip(), inplace=True)
            performance = performance.replace(')','')
            performance.columns = ['Section','time']
            performance = performance.set_index('Section')
            performance[['avg. Time','percentage']] = performance.time.str.split('(',expand=True)
            performance[['%','symbol']] = performance.percentage.str.split(')',expand=True)
            del performance['time']
            del performance['symbol']
            del performance['percentage']

        # Reset file pointer
        log_info.seek(0)
        
        # Return performance data
        return performance

    @property
    def simulations(self):
        """list of dict: parsed data for each separate LAMMPS run/minimize action"""
        return self.__simulations
            
    @property
    def lammps_version(self):
        """str : The LAMMPS version used."""
        return self.__lammps_version
    
    @property
    def lammps_date(self):
        """datetime.date : The date associated with the LAMMPS version used."""
        return self.__lammps_date
    
    def flatten(self, style='last'):
        """
        Combines all simulations into one.  The style options allow for
        duplicate timesteps to be overwritten.
        
        Parameters
        ----------
        style : str, optional
            Specifies which values to use for duplicate time steps:
            - 'first' uses the values from the earliest simulation.
            - 'last' uses the values from the latest simulation (default).
            - 'all' uses all reported lines including ones with duplicate time
              steps.
        """
        # Check that all simulations with thermo data have step values
        for sim in self.simulations:
            if sim.thermo is not None and len(sim.thermo) > 0:
                assert 'Step' in sim.thermo, 'All simulation thermos must have Step key in order to flatten'
        
        # Combine the data into merged_df
        merged_df = self.simulations[0].thermo
        
        for i in range(1, len(self.simulations)):
            if self.simulations[i].thermo is not None:
                thermo = self.simulations[i].thermo
                if style == 'first':
                    merged_df = pd.concat([merged_df, thermo[thermo.Step > merged_df.Step.max()]], ignore_index=True)
                elif style == 'last':
                    merged_df = pd.concat([merged_df[merged_df.Step < thermo.Step.min()], thermo], ignore_index=True)
                elif style == 'all':
                    merged_df = pd.concat([merged_df, thermo], ignore_index=True)
                else:
                    raise ValueError('Unsupported style')
        
        return Simulation(thermo=merged_df)


# coding: utf-8
# cython: language_level=3

# http://cython.org/
import cython

# http://www.numpy.org/
import numpy as np

# atomman imports
from .dmag cimport dmag2_c   

@cython.boundscheck(False)
@cython.wraparound(False)
def nlist(system, double cutoff, Py_ssize_t initialsize=20, Py_ssize_t deltasize=10):
    """
    Calculates a neighbor list for all atoms in a System taking periodic
    boundaries into account.
    
    Parameters
    ----------
    system : atomman.System 
        The system to calculate the neighbor list for.
    cutoff : float
        Radial cutoff distance for identifying neighbors.
    initialsize : int, optional
        The number of neighbor positions to initially assign to each atom.
        Default value is 20.
    deltasize : int, optional
        Specifies the number of extra neighbor positions to allow each atom
        when the number of neighbors exceeds the underlying array size.
        Default value is 10.

    Returns
    -------
    numpy.ndarray of int
        Array listing number of neighbors and neighbor ids for each atom in
        System.  First term in each row is the atom's coordination number, c.
        The next c values are the atom's neighbor ids.  
    """
    
    # Define variables based on input parameters
    pos = system.atoms.pos
    cdef const double[:,:] posv = pos
    cdef const double[:,:] vects = system.box.vects
    cdef const double[:] origin = system.box.origin
    cdef bint pbc_a = system.pbc[0]
    cdef bint pbc_b = system.pbc[1]
    cdef bint pbc_c = system.pbc[2]
    cdef Py_ssize_t maxneighbors = initialsize
    cdef Py_ssize_t maxatomsperbin = 40
    cdef Py_ssize_t natoms = posv.shape[0]
    cdef double cutoff2 = cutoff*cutoff
   
    # Define basic iteration index variables
    cdef Py_ssize_t i, j, k, l

    # Define superbox identification parameters
    cdef Py_ssize_t x, y, z
    cdef double corner
    cdef double[:] supermin=np.empty(3)
    cdef double[:] supermax=np.empty(3)
    
    # Define bins
    cdef double binsize
    cdef double[:] xbins, ybins, zbins
    cdef Py_ssize_t numxbins, numybins, numzbins
    
    # Define bin indexers and associated atom indexes
    #cdef long long [:] xindex, yindex, zindex
    #cdef long long [:,:] xyzindex
    cdef long long [:] atomindex
    cdef long long [:,:] newxyzindex
    cdef long long [:] newatomindex
    
    # Define ghost atom info
    cdef Py_ssize_t xl, xh, yl, yh, zl, zh
    cdef double[:,:] ghostpos = np.empty((0, 3))
    cdef double[:,:] newghostpos
    newpos = np.empty(pos.shape)
    cdef double[:,:] newposv = newpos
    cdef long long[:] ghostindex = np.empty(0, dtype=np.int64)
    cdef long long[:] newindex = np.empty(posv.shape[0], dtype=np.int64)
    cdef long long[:] newghostindex 
    cdef long long [:,:] xyzghostindex
    
    # Define final bins
    cdef Py_ssize_t maxc, c, n
    cdef Py_ssize_t dc, dx, dy, dz
    cdef long long[:, :, :, :] xyzbins, newbins
    
    # Define lists of atoms to compare
    cdef Py_ssize_t uindex, vindex, u, v, w
    cdef long long [:] shortlist, longlist, superlonglist
    cdef bint end
    
    # Define positions and distances between them
    cdef double[:,:] upos, vpos
    cdef double[:] dmag2
    
    # Define neighbors list and assignment terms
    cdef Py_ssize_t uj, vj,
    cdef long long[:, :] neighbors = np.empty((natoms, maxneighbors+1), dtype=np.int64)
    for i in range(natoms):
        neighbors[i, 0] = 0
    cdef long long[:, :] newneighbors
    cdef bint new

    # Determine orthogonal superbox that fully encompasses the system
    for j in range(3):
        supermin[j] = origin[j]
        supermax[j] = origin[j]
    for z in range(0, 2):
        for y in range(0, 2):
            for x in range(0, 2):
                for j in range(3):
                    corner = origin[j] + x * vects[0, j] + y * vects[1, j] + z * vects[2, j]
                    if corner < supermin[j]:
                        supermin[j] = corner
                    if corner > supermax[j]:
                        supermax[j] = corner
    for j in range(3):
        supermin[j] -= 1.01 * cutoff
        supermax[j] += 1.01 * cutoff
    
    # Construct bins
    binsize = cutoff
    xbins = np.arange(supermin[0], supermax[0] + binsize, binsize)
    ybins = np.arange(supermin[1], supermax[1] + binsize, binsize)
    zbins = np.arange(supermin[2], supermax[2] + binsize, binsize)
    numxbins = len(xbins)
    numybins = len(ybins)
    numzbins = len(zbins)
    
    # Build xyz box index for each atom
    xindex = np.digitize(pos[:, 0], xbins) - 1
    yindex = np.digitize(pos[:, 1], ybins) - 1
    zindex = np.digitize(pos[:, 2], zbins) - 1
    xyzindex = np.hstack((xindex[:, np.newaxis], yindex[:, np.newaxis], zindex[:, np.newaxis]))
    
    # Relate atom's id to xyz_index
    atomindex = np.arange(natoms, dtype=np.int64)
    
    # Identify all bins with real atoms
    realbins = unique_rows2(xyzindex)
    
    # Create iterators based on pbc
    if pbc_a:
        xl, xh = -1, 2
    else:
        xl, xh = 0, 1
    if pbc_b:
        yl, yh = -1, 2
    else:
        yl, yh = 0, 1
    if pbc_c:
        zl, zh = -1, 2
    else:
        zl, zh = 0, 1

    # Construct list of ghost atoms in the superbox
    for x in range(xl, xh):
        for y in range(yl, yh):
            for z in range(zl, zh):
                if x == 0 and y == 0 and z == 0:
                    pass
                else:
                    k=0
                    for i in range(posv.shape[0]):
                        for j in range(3):
                            newposv[i, j] = x * vects[0, j] + y * vects[1, j] + z * vects[2, j] + posv[i, j]

                        if (    newposv[i, 0] > supermin[0] and newposv[i, 0] < supermax[0]
                            and newposv[i, 1] > supermin[1] and newposv[i, 1] < supermax[1]
                            and newposv[i, 2] > supermin[2] and newposv[i, 2] < supermax[2]):
                            
                            newindex[k] = i
                            k += 1

                    ghostpos = np.vstack((ghostpos, newpos[newindex[:k]]))
                    ghostindex = np.hstack((ghostindex, newindex[:k]))

    # Append xyzindex and atomindex lists with ghost atoms
    if len(ghostpos) > 0:
        xindex = np.digitize(ghostpos[:, 0], xbins) - 1
        yindex = np.digitize(ghostpos[:, 1], ybins) - 1
        zindex = np.digitize(ghostpos[:, 2], zbins) - 1
        xyzghostindex = np.hstack((xindex[:, np.newaxis], 
                                   yindex[:, np.newaxis],
                                   zindex[:, np.newaxis]))
        xyzindex = np.vstack((xyzindex, xyzghostindex))
        atomindex = np.hstack((atomindex, ghostindex))
    
    # Assign atoms and ghost atoms to xyz bins
    maxc = 0
    xyzbins = np.zeros((numxbins, numybins, numzbins, maxatomsperbin + 1), dtype=np.int64)
    for n in range(atomindex.shape[0]):
        x, y, z = xyzindex[n]
        c = xyzbins[x, y, z, 0] + 1
        
        # Increase size if needed
        if c == maxatomsperbin:
            newbins = np.zeros((numxbins, numybins, numzbins, maxatomsperbin + 11), dtype=np.int64)
            for i in range(xyzbins.shape[0]):
                for j in range(xyzbins.shape[1]):
                    for k in range(xyzbins.shape[2]):
                        for l in range(maxatomsperbin + 1):
                            newbins[i, j, k, l] = xyzbins[i, j, k, l]
            xyzbins = newbins
            maxatomsperbin += 10
        
        if c > maxc:
            maxc = c
        xyzbins[x, y, z, 0] = c
        xyzbins[x, y, z, c] = atomindex[n]

    superlonglist = np.empty(14 * maxc, dtype=np.int64)

    # Iterate over all bins with real atoms
    for i in range(len(realbins)):
        x, y, z = realbins[i]
        c = xyzbins[x, y, z, 0]
        shortlist = np.empty(c, dtype=np.int64)
        
        # shortlist = all atoms in current bin
        for j in range(c):
            shortlist[j] = xyzbins[x, y, z, j+1]
            superlonglist[j] = shortlist[j]
        
        # Add all atoms in half of the nearby bins to longlist
        end = False
        for dz in range(-1, 2):
            for dy in range(-1, 2):
                for dx in range(-1, 2):
                    
                    # Stop when reach center bin
                    if dx == 0 and dy == 0 and dz == 0:
                        end = True
                        break
                    
                    # Skip non-existant neighbor bins
                    if (x + dx < 0 or x + dx == numxbins or
                        y + dy < 0 or y + dy == numybins or
                        z + dz < 0 or z + dz == numzbins):
                        continue
                    
                    dc = xyzbins[x + dx, y + dy, z + dz, 0]
                    for j in range(dc):
                        superlonglist[c+j] = xyzbins[x + dx, y + dy, z + dz, j+1]
                    c += dc
                if end:
                    break
            if end:
                break
                
        longlist = superlonglist[:c]
        
        # Compare all atoms in shortlist to longlist.
        for u in range(shortlist.shape[0]):
            uindex = shortlist[u]
            
            upos = np.empty((longlist.shape[0]-u-1, 3))
            vpos = np.empty((longlist.shape[0]-u-1, 3))

            for w, v in enumerate(range(u+1, longlist.shape[0])):
                for j in range(3):
                    vindex = longlist[v]
                    upos[w, j] = posv[uindex, j]
                    vpos[w, j] = posv[vindex, j]
            
            # Compute distances
            dmag2 = dmag2_c(upos, vpos, vects, pbc_a, pbc_b, pbc_c)
            
            # Assign neighbors if within cutoff
            for w, v in enumerate(range(u+1, longlist.shape[0])):
                if dmag2[w] < cutoff2:
                    vindex = longlist[v]
            
                    if uindex != vindex:
                        new = True
                        uj = -1
                        vj = -1
                        
                        # Find uj position to insert vindex
                        for j in range(1, neighbors[uindex, 0] + 1):
                            # Check uindex's neighbors for vindex
                            if neighbors[uindex, j] == vindex:
                                new = False
                                break
                            elif neighbors[uindex, j] > vindex:
                                uj = j
                                break
                        if uj == -1:
                            uj = neighbors[uindex, 0] + 1

                        if new:
                            # Find vj position to insert uindex
                            for j in range(1, neighbors[vindex, 0] + 1):
                                if neighbors[vindex, j] > uindex:
                                    vj = j
                                    break
                            if vj == -1:
                                vj = neighbors[vindex, 0] + 1

                            # Increase coordination
                            neighbors[uindex, 0] += 1
                            neighbors[vindex, 0] += 1

                            # Extend system size if needed
                            if neighbors[uindex, 0] > maxneighbors or neighbors[vindex, 0] > maxneighbors:
                                newneighbors = np.empty((natoms, maxneighbors + deltasize + 1), dtype=np.int64)
                                for j in range(neighbors.shape[0]):
                                    for k in range(maxneighbors + 1):
                                        newneighbors[j, k] = neighbors[j, k]
                                neighbors = newneighbors
                                maxneighbors += deltasize

                            # Shift neighbor indices with higher values
                            for j in range(neighbors[uindex, 0], uj - 1, -1):
                                neighbors[uindex, j] = neighbors[uindex, j - 1]
                            for j in range(neighbors[vindex, 0], vj - 1, -1):
                                neighbors[vindex, j] = neighbors[vindex, j - 1]

                            # Assign neighbors to each other
                            neighbors[uindex, uj] = vindex
                            neighbors[vindex, vj] = uindex

    return np.asarray(neighbors)

@cython.boundscheck(False)
@cython.wraparound(False)
def unique_rows2(a):  
    """Takes two-dimensional array a and returns only the unique rows."""
    return np.unique(a.view(np.dtype((np.void, a.dtype.itemsize*a.shape[1])))).view(a.dtype).reshape(-1, a.shape[1])


# coding: utf-8

def indexstr(shape):
    """
    Iterates through all unique indicies of an array with a given shape.
    
    Parameters
    ----------
    shape : tuple of int
        The array shape to iterate through.
        
    Yields
    ------
    index : tuple of int
        A unique index set of the array.
    istr : str
        A string representation of index with numbers in [].
    """
    
    if tuple(shape) == ():
        # Yield for empty shape
        yield (), ''
    else:
        # Loop over all values of the first index of shape
        for i in range(shape[0]):
            index1 = (i,)
            istr1 = '['+str(i)+']'
            
            # Recursively go through other indicies of shape
            for index2, istr2 in indexstr(shape[1:]):
                
                # Combine index components
                index = index1 + index2
                istr = istr1 + istr2
                
                # Yield composite
                yield index, istr


# coding: utf-8
from .dvect import dvect # pylint: disable=no-name-in-module
from .dmag import dmag # pylint: disable=no-name-in-module
from .nlist import nlist # pylint: disable=no-name-in-module
from .NeighborList import NeighborList
from .Atoms import Atoms
from .Box import Box
from .ElasticConstants import ElasticConstants
from .System import System
from .displacement import displacement

__all__ = ['displacement', 'dvect', 'dmag', 'nlist', 'Atoms', 'Box',
           'ElasticConstants', 'NeighborList', 'System']


# coding: utf-8
# Standard Python libraries
from copy import deepcopy

# http://www.numpy.org/
import numpy as np

# atomman imports
from . import dislocation_system_transform
from .. import Box
from ..tools import axes_check, vect_angle, miller

class VolterraDislocation(object):
    """
    Generic class for a Volterra solution of a straight dislocation.
    """
    
    def __init__(self, C, burgers, ξ_uvw=None, slip_hkl=None, transform=None,
                 axes=None, box=None, m=[1,0,0], n=[0,1,0], tol=1e-8):
        """
        Initializes the solution. Calls solve.
        
        Parameters
        ----------
        C : atomman.ElasticConstants
            The medium's elastic constants.
        burgers : array-like object
            The dislocation's Burgers vector.
        ξ_uvw : array-like object
            The Miller crystal vector associated with the dislocation's line
            direction.  Must be given with slip_hkl to identify the
            transformation matrix to use on C and burgers.
        slip_hkl : array-like object
            The Miller plane indices associated with the dislocation's slip
            plane.  Must be given with slip_hkl to identify the
            transformation matrix to use on C and burgers.
        transform : array-like object, optional
            A 3x3 set of orthogonal Cartesian vectors that define the
            transformation matrix to use on C and burgers to convert from the
            standard (unit cell) and dislocation orientations.  The 3 vectors
            will automatically be converted into unit vectors.  Using this is
            an alternative to using ξ_uvw and slip_hkl.
        axes : array-like object, optional
            Same as transform.  Retained for backwards compatibility.
        box : atomman.Box, optional
            The unit cell's box that crystal vectors are taken with respect to.
            If not given, will use a cubic box with a=1 meaning that burgers,
            ξ_uvw and slip_hkl will be interpreted as Cartesian vectors.
        m : array-like object, optional
            The m unit vector for the solution.  m, n, and u (dislocation
            line) should be right-hand orthogonal.  Default value is [1,0,0]
            (x-axis).
        n : array-like object, optional
            The n unit vector for the solution.  m, n, and u (dislocation
            line) should be right-hand orthogonal.  Default value is [0,1,0]
            (y-axis). n is normal to the dislocation slip plane.
        tol : float, optional
            Tolerance parameter used to round off near-zero values.  Default
            value is 1e-8.
        """
        self.solve(C, burgers, ξ_uvw=ξ_uvw, slip_hkl=slip_hkl, transform=transform,
                   axes=axes, box=box, m=m, n=n, tol=tol)
        
    def solve(self, C, burgers, ξ_uvw=None, slip_hkl=None, transform=None,
              axes=None, box=None, m=[1,0,0], n=[0,1,0], tol=1e-8):
        """
        Computes parameters for a dislocation solution.
        !!!!GENERIC CLASS ONLY HANDLES INPUT PARAMETERS AND DOES NOT SOLVE!!!!
        
        Parameters
        ----------
        C : atomman.ElasticConstants
            The medium's elastic constants.
        burgers : array-like object
            The dislocation's Burgers vector.
        ξ_uvw : array-like object
            The Miller crystal vector associated with the dislocation's line
            direction.  Must be given with slip_hkl to identify the
            transformation matrix to use on C and burgers.
        slip_hkl : array-like object
            The Miller plane indices associated with the dislocation's slip
            plane.  Must be given with slip_hkl to identify the
            transformation matrix to use on C and burgers.
        transform : array-like object, optional
            A 3x3 set of orthogonal Cartesian vectors that define the
            transformation matrix to use on C and burgers to convert from the
            standard (unit cell) and dislocation orientations.  The 3 vectors
            will automatically be converted into unit vectors.  Using this is
            an alternative to using ξ_uvw and slip_hkl.
        axes : array-like object, optional
            Same as transform.  Retained for backwards compatibility.
        box : atomman.Box, optional
            The unit cell's box that crystal vectors are taken with respect to.
            If not given, will use a cubic box with a=1 meaning that burgers,
            ξ_uvw and slip_hkl will be interpreted as Cartesian vectors.
        m : array-like object, optional
            The m unit vector for the solution.  m, n, and u (dislocation
            line) should be right-hand orthogonal.  Default value is [1,0,0]
            (x-axis).
        n : array-like object, optional
            The n unit vector for the solution.  m, n, and u (dislocation
            line) should be right-hand orthogonal.  Default value is [0,1,0]
            (y-axis). n is normal to the dislocation slip plane.
        tol : float
            Tolerance parameter used to round off near-zero values.  Default
            value is 1e-8.
        """
        # Convert burgers, m, n to numpy arrays if needed
        burgers = np.asarray(burgers, dtype=float)
        m = np.asarray(m, dtype=float)
        n = np.asarray(n, dtype=float)
        
        # Check m, n values
        assert np.isclose(np.linalg.norm(m), 1.0, atol = tol), "m must be a unit vector"
        assert np.isclose(np.linalg.norm(n), 1.0, atol = tol), "n must be a unit vector"
        assert np.isclose(np.dot(m, n), 0.0, atol = tol), "m and n must be perpendicular"
        
        # ξ_uvw-based orientation parameters
        if ξ_uvw is not None or slip_hkl is not None:
            assert ξ_uvw is not None, 'ξ_uvw and slip_hkl must be given together'
            assert slip_hkl is not None, 'ξ_uvw and slip_hkl must be given together'
            assert transform is None, 'transform cannot be given with ξ_uvw, slip_hkl'
            assert axes is None, 'axes cannot be given with ξ_uvw, slip_hkl'
            
            transform = dislocation_system_transform(ξ_uvw, slip_hkl, m=m, n=n, box=box, tol=tol)
        
        # transform-based orientation parameters
        else:
            if axes is not None:
                assert transform is None, 'axes and transform cannot both be given'
                transform = axes
            if transform is not None:
                transform = axes_check(transform)
            else:
                transform = np.eye(3, dtype=float)

        # Set default box
        if box is None:
            box = Box()
        
        # Transform burgers and C
        burgers = miller.vector_crystal_to_cartesian(burgers, box)
        burgers = transform.dot(burgers)
        burgers[np.isclose(burgers / np.abs(burgers).max(), 0.0, atol = tol)] = 0.0
        C = C.transform(transform)
        
        self.__C = C
        self.__m = m
        self.__n = n
        self.__ξ = np.cross(m, n)
        self.__burgers = burgers
        self.__tol = tol
        self.__transform = transform
    
    @property
    def m(self):
        return self.__m
    
    @property
    def n(self):
        return self.__n
    
    @property
    def ξ(self):
        return self.__ξ
    
    @property
    def C(self):
        return self.__C
    
    @property
    def burgers(self):
        return self.__burgers
    
    @property
    def tol(self):
        return self.__tol
    
    @property
    def transform(self):
        return self.__transform

    def characterangle(self, unit='degree'):
        """
        Returns the dislocation's character angle.
        
        Parameters
        ----------
        unit : str, optional
            Specify whether the angle is given in 'degree' (default)
            or 'radian'.
            
        Returns
        -------
        float
            The dislocation character angle.
        """
        return vect_angle(self.burgers, self.ξ, unit=unit)
    
    @property
    def K_coeff(self):
        """float : The energy coefficient"""
        
        # K = b_i K_ij b_j / (b_k b_k)
        return (self.burgers.dot(self.K_tensor.dot(self.burgers))
                / self.burgers.dot(self.burgers))
    
    @property
    def K_tensor(self):
        raise NotImplementedError('Needs to be defined by subclass')
    
    @property
    def preln(self):
        """float : The pre-ln strain energy factor"""
        # a = b_i K_ij b_j / (4 π)
        return self.burgers.dot(self.K_tensor.dot(self.burgers)) / (4 * np.pi)
    
    def displacement(self, pos):
        raise NotImplementedError('Needs to be defined by subclass')
    
    def stress(self, pos):
        raise NotImplementedError('Needs to be defined by subclass')


# coding: utf-8
# cython: language_level=3
# Standard Python libraries

# http://cython.org/
import cython

# http://www.numpy.org/
import numpy as np

# atomman imports
from .. import NeighborList
from ..core.dvect cimport dvect_c

@cython.boundscheck(False)
@cython.wraparound(False)
def slip_vector(system_0, system_1, neighbors=None, cutoff=None):
    """
    Compute the slip vectors for all atoms.  Note that this differs from the
    original formulation in that it is not normalized by number of slipped
    neighbors.
    
        s_i = -Σ_j d_ij(t) - d_ij(0)
    
    where j is neighbor atoms of atom i, and d_ij() is vector distance between
    atoms i and j at time t.
    
    Parameters
    ----------
    system_0 : atomman.system
        The base/reference system to use.
    system_1 : atomman.system
        The defect/current system to use.
    neighbors : atomman.NeighborList, optional
        The neighbor list associated with system_0 to use.  Either neighbors
        or cutoff must be given, or system_0 must have a neighbors attribute.
    cutoff : float, optional
        Cutoff distance for computing a neighbor list for system_0.  Either
        neighbors or cutoff must be given, or system_0 have a neighbors
        attribute.
    
    Returns
    -------
    numpy.ndarray
        The computed slip vectors.
    """
    # Check that the two systems have the same number of atoms
    if system_0.natoms != system_1.natoms:
        raise ValueError('systems have different number of atoms')

    # Neighbor list setup
    if neighbors is not None:
        assert cutoff is None, 'neighbors and cutoff cannot both be given'
    elif cutoff is not None:
        neighbors = NeighborList(system=system_0, cutoff=cutoff)
    elif hasattr(system_0, 'neighbors'):
        neighbors = system_0.neighbors
    else:
        raise ValueError('neighbors or cutoff is required')
    
    pos_0 = system_0.atoms.pos
    pos_1 = system_1.atoms.pos
    bvects = system_0.box.vects
    nlist = neighbors.nlist
    
    return slip_vector_c(pos_0, pos_1, bvects, nlist, system_0.pbc[0], system_0.pbc[1], system_0.pbc[2])

@cython.boundscheck(False)
@cython.wraparound(False)
cdef slip_vector_c(const double[:,:] pos_0, 
                   const double[:,:] pos_1, 
                   const double[:,:] bvects, 
                   const long long [:,:] nlist, 
                   bint pbc_a, bint pbc_b, bint pbc_c):
    
    cdef Py_ssize_t i, j, coordmax, ni, n, coord
    
    cdef double[:,:] d_0, d_1
    
    # Define the slip vector
    slip = np.zeros((pos_0.shape[0], 3))
    cdef double[:,:] slipv = slip
    
    coordmax = 0
    for i in range(nlist.shape[0]):
        if nlist[i, 0] > coordmax:
            coordmax = nlist[i, 0]
    
    cdef double[:,:] upos_0, vpos_0, upos_1, vpos_1
    upos_0 = np.zeros((coordmax, 3))
    vpos_0 = np.zeros((coordmax, 3))
    upos_1 = np.zeros((coordmax, 3))
    vpos_1 = np.zeros((coordmax, 3))
    
    for i in range(pos_0.shape[0]):
        coord = nlist[i, 0]
        for n in range(coord):
            for j in range(3):
                ni = nlist[i, n+1]
                upos_0[n, j] = pos_0[i, j]
                vpos_0[n, j] = pos_0[ni, j]
                upos_1[n, j] = pos_1[i, j]
                vpos_1[n, j] = pos_1[ni, j]
                
        d_0 = dvect_c(upos_0, vpos_0, bvects, pbc_a, pbc_b, pbc_c)
        d_1 = dvect_c(upos_1, vpos_1, bvects, pbc_a, pbc_b, pbc_c)
        
        for n in range(coord):
            for j in range(3):
                slipv[i, j] -= d_1[n, j] - d_0[n, j]
    
    return slip


# coding: utf-8

import matplotlib.pyplot as plt

# http://www.numpy.org/
import numpy as np

from . import gradient, integrator
import atomman.unitconvert as uc

class BasePath():
    """
    Generic class representing an energy path.
    """
    
    def __init__(self, coord, energyfxn, gradientfxn='cdiff',
                gradientkwargs=None, integratorfxn='rk'):
        """
        Class initializer.
        
        Parameters
        ----------
        coord : array-like object
            The list of coordinates associated with the points along the path.
        energyfxn : function
            The function that evaluates the energy associated with the different
            point coordinates.
        gradientfxn : str or function, optional
            The function to use to estimate the gradient of the energy.  Default
            value of 'cdiff' will use atomman.mep.gradient.central_difference
        gradientkwargs : dict, optional
            The keyword arguments (i.e. settings) to use with the gradientfxn.
            Default is an empty dictionary, i.e. default settings of gradientfxn.
        integratorfxn : str or function, optional
            The function to use to integrate relaxation steps.  Default value of
            'rk' will use atomman.mep.integrator.rungekutta.
        """
        
        if isinstance(coord, BasePath):
            coord = coord.coord
        self.coord = coord

        if callable(energyfxn):
            self.__energyfxn = energyfxn
        else:
            raise TypeError('energyfxn must be a callable object')
        
        self.gradientfxn = gradientfxn
        self.integratorfxn = integratorfxn
        
        if gradientkwargs is None:
            self.__gradientkwargs = {}
        if isinstance(gradientkwargs, dict):
            self.__gradientkwargs = gradientkwargs
        else:
            raise TypeError('gradientkwargs must be None or a dict')
    
    @property
    def coord(self):
        """numpy.NDArray : The coordinates for each point along the path."""
        return self.__coord
    
    @coord.setter
    def coord(self, value):
        self.__coord = np.asarray(value)
    
    @property
    def arccoord(self):
        """numpy.NDArray : The arc length coordinates for each point along the path."""
        
        # Compute the arc length distance between each point
        s = np.zeros(len(self.coord))
        s[1:] = np.linalg.norm(self.coord[1:] - self.coord[:-1], axis=1)
        
        # Compute the arc length coordinates
        α = np.empty(len(self.coord))
        for i in range(len(α)):
            α[i] = s[:i+1].sum()
        
        return α

    @property
    def energyfxn(self):
        """function : The function for evaluating the energy"""
        return self.__energyfxn
    
    @property
    def gradientfxn(self):
        """function : The function for evaluating the energy gradients"""
        return self.__gradientfxn
    
    @gradientfxn.setter
    def gradientfxn(self, value):
        if isinstance(value, str):
            if value == 'central_difference' or value == 'cdiff':
                self.__gradientfxn = gradient.central_difference
            else:
                raise ValueError('Unknown gradientfxn style')
        elif callable(value):
            self.__gradientfxn = value
        else:
            raise TypeError('gradientfxn must be a str of a known style or a callable object')

    @property
    def gradientkwargs(self):
        """dict : The keyword arguments to use when calling gradientfxn"""
        return self.__gradientkwargs
    
    @property
    def integratorfxn(self):
        """function : The function to use for integrating minimization steps"""
        return self.__integratorfxn
    
    @integratorfxn.setter
    def integratorfxn(self, value):
        if isinstance(value, str):
            if value == 'rungekutta' or value == 'rk':
                self.__integratorfxn = integrator.rungekutta
            elif value == 'euler':
                self.__integratorfxn = integrator.euler
            else:
                raise ValueError('Unknown integratorfxn style')
        elif callable(value):
            self.__integratorfxn = value
        else:
            raise TypeError('integratorfxn must be a str of a known style or a callable object')

    @property
    def unittangent(self):
        """numpy.NDArray : The tangent vectors along the path at each point."""
        raise NotImplementedError('Defined by subclasses')
    
    def energy(self, coord=None):
        """
        Evaluates energy values associated with either the path points or
        given coordinates.
        
        Parameters
        ----------
        coord : array-like object, optional
            Coordinates to evaluate the energies at.  If not given, will use
            the object's coord values.
            
        Returns
        -------
        np.NDArray
            The evaluated energies.
        """
        if coord is None:
             coord = self.coord
        
        return self.energyfxn(coord)
    
    def grad_energy(self, coord=None):
        """
        Evaluates the gradient of the energy associated with either the path
        points or given coordinates.
        
        Parameters
        ----------
        coord : array-like object, optional
            Coordinates to evaluate the energy gradients at.  If not given,
            will use the object's coord values.
            
        Returns
        -------
        np.NDArray
            The evaluated energies.
        """
        if coord is None:
             coord = self.coord
        return self.gradientfxn(self.energyfxn, coord, **self.gradientkwargs) #pylint: disable=not-callable
    
    @property
    def force(self):
        """numpy.NDArray : The computed force associated with moving along the path at each point."""
        return np.einsum('ij,ij->i', self.grad_energy(), self.unittangent)

    def step(self, *args, **kwargs):
        """
        Performs a single relaxation step.
        
        Returns
        -------
        newpath : Path
            A Path with coordinates evolved forward by one timestep.
        """
        raise NotImplementedError('Defined by subclasses')

    def relax(self, *args, **kwargs):
        """
        Perform multiple relaxation and/or climb steps until either the
        maximum coordinate displacement per step drops below a tolerance or
        the maximum number of steps is reached.

        Returns
        -------
        newpath : Path
            A Path with coordinates evolved forward from the relaxation.
        """
        raise NotImplementedError('Defined by subclasses')

    def plot_energy(self, energy_unit=None, length_unit=None, ax=None, **kwargs):
        """
        Creates a plot of the energies along the path as a function of the
        arc coordinates.

        Parameters
        ----------
        energy_unit : str or None, optional
            If given, the energy values will be converted from atomman's
            working units to the specified units.  Default value of None will
            do no conversions, which is useful if the energyfxn is not reporting
            in atomman's working units.
        length_unit : str or None, optional
            If given, the arc coordinates will be converted from atomman's
            working units to the specified units.  Default value of None will
            do no conversions, which is useful if the coords are not given
            in atomman's working units.
        ax : matplotlib.pyplot.axis
            A pre-existing plotting axis.  Allows for more control and the
            use of subplots.
        **kwargs : any, optional
            All additional keyword arguments will be passed to 
            matplotlib.pyplot.figure().

        Returns
        -------
        matplotlib.pyplot.figure
            The generated figure allowing for further modifications.  Returned
            if ax is None.
        """
        if ax is None:
            fig = plt.figure(**kwargs)
            ax = fig.add_subplot(111)
            returnfig = True
        else:
            assert len(kwargs) == 0, 'ax and extra kwargs cannot both be given'
            returnfig = False

        # Handle energy units
        energy = self.energy()
        if energy_unit is None:
            ax.set_ylabel('Energy', size='x-large')
        else:
            ax.set_ylabel(f'Energy ({energy_unit})', size='x-large')
            energy = uc.get_in_units(energy, energy_unit)

        # Handle length unit
        arccoord = self.arccoord
        if length_unit is None:
            ax.set_xlabel('Arc coordinate', size='x-large')
        else:
            ax.set_xlabel(f'Arc coordinate ({length_unit})', size='x-large')
            arccoord = uc.get_in_units(arccoord, length_unit)
        ax.set_xlim(0, arccoord[-1])

        ax.plot(arccoord, energy, lw=3)

        if returnfig:
            return fig

    def plot_force(self, force_unit=None, length_unit=None, ax=None, **kwargs):
        """
        Creates a plot of the force to move along the path as a function of the
        arc coordinates.

        Parameters
        ----------
        force_unit : str or None, optional
            If given, the force values will be converted from atomman's
            working units to the specified units.  Default value of None will
            do no conversions, which is useful if the energyfxn is not reporting
            in atomman's working units.
        length_unit : str or None, optional
            If given, the arc coordinates will be converted from atomman's
            working units to the specified units.  Default value of None will
            do no conversions, which is useful if the coords are not given
            in atomman's working units.
        ax : matplotlib.pyplot.axis
            A pre-existing plotting axis.  Allows for more control and the
            use of subplots.
        **kwargs : any, optional
            All additional keyword arguments will be passed to 
            matplotlib.pyplot.figure().

        Returns
        -------
        matplotlib.pyplot.figure
            The generated figure allowing for further modifications.  Returned
            if ax is None.
        """
        if ax is None:
            fig = plt.figure(**kwargs)
            ax = fig.add_subplot(111)
            returnfig = True
        else:
            assert len(kwargs) == 0, 'ax and extra kwargs cannot both be given'
            returnfig = False

        # Handle force units
        force = self.force
        if force_unit is None:
            ax.set_ylabel('Force', size='x-large')
        else:
            ax.set_ylabel(f'Force ({force_unit})', size='x-large')
            force = uc.get_in_units(force, force_unit)

        # Handle length unit
        arccoord = self.arccoord
        if length_unit is None:
            ax.set_xlabel('Arc coordinate', size='x-large')
        else:
            ax.set_xlabel(f'Arc coordinate ({length_unit})', size='x-large')
            arccoord = uc.get_in_units(arccoord, length_unit)
        ax.set_xlim(0, arccoord[-1])

        ax.plot(arccoord, force, lw=3)

        if returnfig:
            return fig


# coding: utf-8

# http://www.numpy.org/
import numpy as np

# atomman imports
from .. import NeighborList

def nye_tensor_p(system, neighbors=None, cutoff=None):
    """
    Generates a list of p vectors for each atom to be used by the nye_tensor()
    function.  The p vectors correspond to the radial distance vectors between
    an atom and each of its neighbors within a cutoff distance in a "perfect
    crystal" reference state.
    
    Parameters
    ----------
    system : atomman.system
        The base/reference system to use.  This should be a defect-free
        perfect crystal system with atom ids directly corresponding to atoms
        in any system that you want to analyze with the Nye tensor.
    neighbors : atomman.NeighborList, optional
        The neighbor list associated with system to use.  Either neighbors
        or cutoff must be given, or system must have a neighbors attribute.
    cutoff : float
        Cutoff distance for computing a neighbor list for system.  Either
        neighbors or cutoff must be given, or system have a neighbors
        attribute.
    
    Returns
    -------
    numpy.ndarray
        The list of p distance vectors for each atom in system.
    """
    
    # Neighbor list setup
    if neighbors is not None:
        assert cutoff is None, 'neighbors and cutoff cannot both be given'
    elif cutoff is not None:
        neighbors = NeighborList(system=system, cutoff=cutoff)
    elif hasattr(system, 'neighbors'):
        neighbors = system.neighbors
    else:
        raise ValueError('neighbors or cutoff is required')
    
    p = []
    for i in range(system.natoms):
        p.append(system.dvect(i, neighbors[i]))
    
    return np.asarray(p)


import numpy as np

from . import free_surface_basis
from ..tools.crystalsystem import iscubic
from ..tools import miller

class FreeSurface():
    """
    Class for generating free surface atomic configurations using clean planar slices. 
    """
    
    def __init__(self, hkl, ucell, cutboxvector='c', maxindex=None,
                 conventional_setting='p', tol=1e-7):
        """
        Class initializer.  Identifies the proper rotations for the given hkl plane
        and cutboxvector, and creates the rotated cell.
        
        Parameters
        ----------
        hkl : array-like object
            The free surface plane to generate expressed in either 3 indices
            Miller (hkl) format or 4 indices Miller-Bravais (hkil) format.
        ucell : atomman.System
            The unit cell to use in generating the system.
        cutboxvector : str, optional
            Specifies which of the three box vectors corresponds to the
            out-of-plane vector.  Default value is c.
        maxindex : int, optional
            Max uvw index value to use in identifying the best uvw set for the
            out-of-plane vector.  If not given, will use the largest absolute
            index between the given hkl and the initial in-plane vector guesses.
        conventional_setting : str, optional
            Allows for rotations of a primitive unit cell to be determined from
            (hkl) indices specified relative to a conventional unit cell.  Allowed
            settings: 'p' for primitive (no conversion), 'f' for face-centered,
            'i' for body-centered, and 'a', 'b', or 'c' for side-centered.  Default
            behavior is to perform no conversion, i.e. take (hkl) relative to the
            given ucell.
        tol : float, optional
            Tolerance parameter used to round off near-zero values.  Default
            value is 1e-8.
        """
        
        # Pass parameters to free_surface_basis to get the rotation uvws
        uvws = free_surface_basis(hkl, box=ucell.box, cutboxvector=cutboxvector, maxindex=maxindex,
                                  conventional_setting=conventional_setting)
        
        # Generate the rotated cell
        rcell, transform = ucell.rotate(uvws, return_transform=True)
        
        # Transform uvws to the conventional cell representation
        if uvws.shape == (3,3):
            uvws = miller.vector_primitive_to_conventional(uvws, conventional_setting)

        # Set cutindex and rcellwidth based on cutboxvector
        if cutboxvector == 'a':
            if rcell.box.bvect[0] != 0.0 or rcell.box.cvect[0] != 0.0:
                raise ValueError("box bvect and cvect cannot have x component for cutboxvector='a'")
            cutindex = 0
        
        elif cutboxvector == 'b':
            if rcell.box.avect[1] != 0.0 or rcell.box.cvect[1] != 0.0:
                raise ValueError("box avect and cvect cannot have y component for cutboxvector='b'")
            cutindex = 1

        elif cutboxvector == 'c':
            if rcell.box.avect[2] != 0.0 or rcell.box.bvect[2] != 0.0:
                raise ValueError("box avect and bvect cannot have z component for cutboxvector='c'")
            cutindex = 2
        
        # Define out of plane unit vector 
        ovect = np.zeros(3)
        ovect[cutindex] = 1.0
        
        # Get out of plane width
        rcellwidth = rcell.box.vects[cutindex, cutindex]
        
        # Get the unique coordinates normal to the plane
        pos = rcell.atoms.pos
        numdec = - int(np.floor(np.log10(tol)))
        coords = np.unique(pos[:, cutindex].round(numdec))

        # Add periodic replica if missing
        if not np.isclose(coords[-1] - coords[0], rcellwidth, rtol=0.0, atol=tol):
            coords = np.append(coords, coords[0] + rcellwidth)
            
        # Compute the shifts
        relshifts = rcellwidth - (coords[1:] + coords[:-1]) / 2
        relshifts[relshifts > rcellwidth] -= rcellwidth
        relshifts[relshifts < 0.0] += rcellwidth
        shifts = np.outer(np.sort(relshifts), ovect)
        
        # Save attributes
        self.__hkl = np.asarray(hkl)
        self.__ucell = ucell
        self.__rcell = rcell
        self.__cutboxvector = cutboxvector
        self.__cutindex = cutindex
        self.__uvws = uvws
        self.__rcellwidth = rcellwidth
        self.__shifts = shifts
        self.__transform = transform
        self.__conventional_setting = conventional_setting

    @property
    def hkl(self):
        """list : Crystal plane in Miller or Miller-Bravais indices"""
        return self.__hkl
    
    @property
    def ucell(self):
        """atomman.System : The unit cell to use in building the defect system."""
        return self.__ucell
    
    @property
    def rcell(self):
        """atomman.System : the rotated cell to use in building the defect system."""
        return self.__rcell
    
    @property
    def cutboxvector(self):
        """str : The box vector for the cut direction."""
        return self.__cutboxvector
    
    @property
    def cutindex(self):
        """int : The Cartesian index for the cut direction."""
        return self.__cutindex
    
    @property
    def uvws(self):
        """numpy.ndarray : The conventional Miller or Miller-Bravais crystal vectors associated with the rcell box vectors."""
        return self.__uvws
    
    @property
    def rcellwidth(self):
        """float : The width of rcell in the cutindex direction."""
        return self.__rcellwidth
    
    @property
    def shifts(self):
        """list : All shift values that place the fault halfway between atomic layers in rcell."""
        return self.__shifts
    
    @property
    def system(self):
        """atomman.System : The built free surface system."""
        try:
            return self.__system
        except:
            raise AttributeError('system not yet built. Use build_system() or surface().')
        
    @property
    def surfacearea(self):
        """float : The surface area of one of the hkl planes."""
        try:
            return self.__surfacearea
        except:
            raise AttributeError('system not yet built. Use build_system() or surface().')
        
    @property
    def transform(self):
        """numpy.ndarray : The Cartesian transformation tensor associated with rotating from ucell to rcell"""
        return self.__transform
    
    @property
    def conventional_setting(self):
        """str : The lattice setting/centering associated with the conventional cell (used if ucell is primitive)"""
        return self.__conventional_setting

    def surface(self, shift=None, vacuumwidth=None, minwidth=None, sizemults=None,
                even=False):
        """
        Generates and returns a free surface atomic system.
        
        Parameters
        ----------
        shift : array-like object, optional
            Applies a shift to all atoms. Different values allow for free surfaces with
            different termination planes to be selected.
        vacuumwidth : float, optional
            If given, the free surface is created by modifying the system's box to insert
            a region of vacuum with this width. This is typically used for DFT calculations
            where it is computationally preferable to insert a vacuum region and keep all
            dimensions periodic.
        sizemults : list or tuple, optional
            The three System.supersize multipliers [a_mult, b_mult, c_mult] to use on the
            rotated cell to build the final system. Note that the cutboxvector sizemult
            must be an integer and not a tuple.  Default value is [1, 1, 1].
        minwidth : float, optional
            If given, the sizemult along the cutboxvector will be selected such that the
            width of the resulting final system in that direction will be at least this
            value. If both sizemults and minwidth are given, then the larger of the two
            in the cutboxvector direction will be used. 
        even : bool, optional
            A True value means that the sizemult for cutboxvector will be made an even
            number by adding 1 if it is odd.  Default value is False.       

        Returns
        -------
        atomman.System
            The free surface atomic system.
        """

        # Set default function values
        if shift is None:
            shift = np.zeros(3)
        else:
            shift = np.asarray(shift)
        
        if sizemults is None:
            sizemults = [1, 1, 1]
        
        # Handle minwidth
        if minwidth is not None:
            mult = int(np.ceil(minwidth / self.rcellwidth))
            
            sizemult = sizemults[self.cutindex]
            if mult > np.abs(sizemult):
                sizemults[self.cutindex] = np.sign(sizemult) * mult
        
        # Handle even
        if even and sizemults[self.cutindex] % 2 == 1:
            if sizemults[self.cutindex] > 0:
                sizemults[self.cutindex] += 1
            else:
                sizemults[self.cutindex] -= 1
        
        # Define out of plane unit vector 
        ovect = np.zeros(3)
        ovect[self.cutindex] = 1.0

        # Supersize and shift the system
        system = self.rcell.supersize(*sizemults)
        system.atoms.pos += shift
        system.wrap()
        
        # Change system's pbc
        system.pbc = [True, True, True]
        system.pbc[self.cutindex] = False
        
        # Insert vacuumwidth
        if vacuumwidth is not None:
            if vacuumwidth < 0:
                raise ValueError('vacuumwidth must be positive')
            newvects = system.box.vects
            newvects[self.cutindex, self.cutindex] += vacuumwidth
            neworigin = system.box.origin - ovect * vacuumwidth / 2 
            system.box_set(vects=newvects, origin=neworigin)
               
        # Compute surfacearea based on cutboxvector
        if self.cutboxvector == 'a':
            surfacearea = np.linalg.norm(np.cross(system.box.bvect, system.box.cvect))

        elif self.cutboxvector == 'b':
            surfacearea = np.linalg.norm(np.cross(system.box.avect, system.box.cvect))

        elif self.cutboxvector == 'c':
            surfacearea = np.linalg.norm(np.cross(system.box.avect, system.box.bvect))
                        
        # Save attributes
        self.__system = system
        self.__surfacearea = surfacearea

        return self.system


# coding: utf-8

# http://www.numpy.org/
import numpy as np

# http://pymatgen.org
try:
    import pymatgen as pmg
    has_pmg = True
except:
    has_pmg = False

def dump(system, symbols=None):
    """
    Convert an atomman.System into a pymatgen.Structure.
    
    Parameters
    ----------
    system : atomman.System
        A atomman representation of a system.
    symbols : tuple, optional
        List of the element symbols that correspond to the atom types.  If not
        given, will use system.symbols if set, otherwise no element content
        will be included.
    
    Returns
    -------
    structure : pymatgen.Structure
        A pymatgen representation of a structure.
    """
    
    assert has_pmg, 'pymatgen not imported'
    
    # Get box/lattice information
    lattice = pmg.Lattice(system.box.vects)
    
    # Get symbols information
    if symbols is None:
        symbols = system.symbols
    symbols = np.asarray(symbols)
    if None in symbols:
        raise ValueError('Symbols needed for all atypes')
    
    # Convert short symbols list to full species list
    atype = system.atoms.atype
    species = symbols[atype-1]
    
    # Get atomic information
    sites = system.atoms_prop(key='pos', scale=True)
    prop = {}
    for p in system.atoms_prop():
        if p != 'atype' and p != 'pos':
            prop[p] = system.atoms_prop(key=p)
    
    # Build structure
    structure = pmg.Structure(lattice, species, sites, site_properties=prop)
    
    return structure


# coding: utf-8
# Standard Python libraries
import os

# http://www.numpy.org/
import numpy as np

# http://www.diffpy.org/diffpy.structure/
try:
    import diffpy.structure
    has_diffpy = True
except:
    has_diffpy = False

# atomman imports
from ... import Atoms, Box, System
from ...tools import uber_open_rmode

def load(cif, symbols=None):
    """
    Reads in a CIF crystal file.  Requires diffpy.Structure.
    
    Parameters
    ----------
    cif : str or file-like object
        The cif content to read.
    symbols : tuple, optional
        Allows the list of element symbols to be assigned during loading.
        Useful if the symbols for the model differ from the standard element
        tags.
    
    Returns
    -------
    system : atomman.System
        An atomman representation of a system.
    """
    
    assert has_diffpy, 'diffpy.Structure not imported'
    
    # Read in cif file to diffpy Structure
    dps = diffpy.structure.structure.Structure()
    with uber_open_rmode(cif) as f:
        dps.readStr(f.read().decode('UTF-8'))
    
    all_elements = dps.element
    elements, all_atype = np.unique(all_elements, return_inverse=True)
    all_atype += 1
    all_pos = dps.xyz
    
    atype = []
    pos = []
    for a1, p1 in zip(all_atype, all_pos):
        noMatch = True
        for a2, p2 in zip(atype, pos):
            if a1 == a2 and np.allclose(p1, p2):
                noMatch = False
                break
        if noMatch:
            atype.append(a1)
            pos.append(p1)
    
    atoms = Atoms(atype=atype, pos=pos)
    
    box = Box(a=dps.lattice.a,
              b=dps.lattice.b,
              c=dps.lattice.c,
              alpha=dps.lattice.alpha,
              beta=dps.lattice.beta,
              gamma=dps.lattice.gamma)
    
    return System(atoms=atoms, box=box, scale=True, symbols = elements)


# coding: utf-8

# http://www.numpy.org/
import numpy as np

from . import Shape

class Sphere(Shape):
    """
    Class representing a sphere in space.
    """
    def __init__(self, center, radius):
        """
        Defines a sphere

        Parameters
        ----------
        center : array-like object
            The position of the sphere's center.
        radius : float
            The sphere's radius. 
        """
        self.center = center
        self.radius = radius
    
    @property 
    def center(self):
        """numpy.NDArray : 3D vector position of sphere's center"""
        return self.__center
    
    @center.setter
    def center(self, value):
        # Check that value is proper
        value = np.asarray(value)
        assert value.shape == (3,), "center must be a 3D vector"
        
        # Save
        self.__center = value
    
    @property
    def radius(self):
        """float : the sphere's radius"""
        return self.__radius

    @radius.setter
    def radius(self, value):
        # Check that value is proper
        value = float(value)
        assert value > 0, "radius must be positive"

        # Save
        self.__radius = value
        
    def inside(self, pos, inclusive=True):
        """
        Indicates if position(s) are inside the shape.
        
        Parameters
        ----------
        pos : array-like object
            Nx3 array of coordinates. 
        inclusive : bool, optional
            Indicates if points on the shape's boundaries are to be included.
            Default value is True.
        
        Returns
        -------
        numpy.NDArray
            N array of bool values: True if inside shape
        """
        
        # Ensure pos is a numpy array 
        pos = np.asarray(pos)
        
        # Find distance of pos from center
        rpos = np.linalg.norm(pos - self.center, axis=-1)
        
        if inclusive:
            return rpos <= self.radius
        else:
            return rpos < self.radius


# coding: utf-8
# Standard Python libraries
from copy import deepcopy

# http://www.numpy.org/
import numpy as np

# atomman imports
from ...tools.crystalsystem import identifyfamily

#Generates the LAMMPS script lines associated with having LAMMPS create the system
def dump(system, atom_style=None, units=None, natypes=None, 
         potential=None, uvws=None, shift=None, size_mults=None,
         return_pair_info=False, safecopy=False):
    """
    Write LAMMPS commands that result in LAMMPS generating an atomic configuration
    based on the supplied system.  
    
    NOTE: It is recommended that this only be used for small systems (e.g. unit cells)
    and to instead use  'atom_data' dump style for more complex configurations.
    
    Parameters
    ----------
    system : atomman.System 
        The system that LAMMPS will replicate.
    atom_style : str, optional
        The LAMMPS atom_style option associated with the data file.  If neither
        atom_style or potential is given, will set atom_style to 'atomic'.
    units : str, optional
        The LAMMPS units option associated with the data file.  If neither
        units or potential is given, will set units 'metal'.
    natypes : int, optional
        Allows the natypes value to be manually changed.  This is needed if
        natypes needs to be greater than the current number of atypes.  If
        neither natypes or potential is given, will use system.natypes.
    potential : atomman.lammps.Potential, optional
        Potential-specific values of atom_style, units, and natypes can be
        extracted from a Potential object.  If both potential and any of the
        individual values are given, the individual values will be used.
    uvws : array-like object, optional
        3x3 array of integer Miller crystal vectors to reorient the system along.
        Currently only supported for cubic systems.
    shift : array-like object, optional
        A 3D rigid-body shift vector to apply to the atomic positions.  Useful to
        avoid positioning atoms directly on the boundaries.  Default value is 
        [0.01, 0.01, 0.01].
    size_mults : array-like object, optional
        3 sets of (-,+) size multipliers to apply to the system.  If not given, no
        multiplications of the system will be made.
    return_pair_info : bool, optional
        Indicates if the LAMMPS command lines associated with setting mass,
        pair_style and pair_coeff are included in the returned info.  If True,
        potential must be given.  Default value is False.
    safecopy : bool, optional
        The LAMMPS commands requires all atoms to be inside box bounds, i.e.
        "wrapped".  If safecopy is True then a copy of the system is made to
        keep the original unwrapped.  Default value is False.
    
    Returns
    -------
        The LAMMPS input command lines to generate the system for itself.
        If return_pair_info is True and potential is given, the LAMMPS input
        command lines for the potential will also be included.

    Raises
    ------
    ValueError
        If return_pair_info is True and potential is not given.
    ValueError
        If the given system is not normalized for LAMMPS compatibility.
    ValueError
        If rotation axes are given for a non-orthogonal system.
    """
    
    # Test that box parameters are compatible with LAMMPS
    if not system.box.is_lammps_norm():
        raise ValueError('System not normalized for LAMMPS compatibility')
    
    # Add comment line
    info = '# Script prepared using atomman Python package\n\n'
    
    # Extract potential-based parameters
    if potential is not None:
        if units is None:
            units = potential.units
        if atom_style is None:
            atom_style = potential.atom_style
        if natypes is None:
            natypes = len(potential.normalize_symbols(system.symbols))
    
    # Set default parameter values
    else:
        if units is None:
            units = 'metal'
        if atom_style is None:
            atom_style = 'atomic'
        if natypes is None:
            natypes = system.natypes
    
    # Generate units and atom_style lines
    info += f'units {units}\n'
    info += f'atom_style {atom_style}\n\n'
    
    # Wrap atoms
    if safecopy:
        system = deepcopy(system)
    system.wrap()
    
    family = identifyfamily(system.box)
    
    # Generate boundary condition
    info += 'boundary '
    for pbc in system.pbc:
        if pbc:
            info += 'p '
        else:
            info += 'm '
    info += '\n\n'
    
    # Start lattice command with box parameters
    vects = system.box.vects
    info += 'lattice custom 1.0 &\n'
    info += f'        a1 {vects[0,0]:.12f} {vects[0,1]:.12f} {vects[0,2]:.12f} &\n'
    info += f'        a2 {vects[1,0]:.12f} {vects[1,1]:.12f} {vects[1,2]:.12f} &\n'
    info += f'        a3 {vects[2,0]:.12f} {vects[2,1]:.12f} {vects[2,2]:.12f} &\n'
    
    # Add origin shift
    if shift is None:
        shift = np.array([0.01, 0.01, 0.01])
    info += f'        origin {shift[0]} {shift[1]} {shift[2]} &\n'
    
    # Add spacing and orient
    if uvws is not None:
        if family not in ['cubic', 'trigonal', 'orthorhombic']:
            raise ValueError('uvws not currently supported for non-orthorhombic systems')
         
        # Adjust crystal spacing for systems to be (nearly) perfectly periodic across boundaries
        spacing = np.zeros(3)
        for i in range(3):
            spacing[i] = vects[i,i] * np.linalg.norm(uvws[i])
            
        info += f'        spacing {spacing[0]:.12f} {spacing[1]:.12f} {spacing[2]:.12f}&\n'
        info += f'        orient x {uvws[0,0]} {uvws[0,1]} {uvws[0,2]} &\n'
        info += f'        orient y {uvws[1,0]} {uvws[1,1]} {uvws[1,2]} &\n'
        info += f'        orient z {uvws[2,0]} {uvws[2,1]} {uvws[2,2]} &\n'
    
    # Add basis (i.e. positions)
    spos = system.atoms_prop('pos', scale=True)
    for i in range(system.natoms):
        info += f'        basis {spos[i,0]:.12f} {spos[i,1]:.12f} {spos[i,2]:.12f}'
        if i < system.natoms - 1:
            info += ' &\n'
    info += '\n\n'
    
    # Generate region
    # Identify if system is orthorhombic or not
    if family in ['cubic', 'trigonal', 'orthorhombic']:
        region_style = 'block'
    else:
        region_style = 'prism'
    
    # Generate the system region 
    if size_mults is None:
        size_mults = np.array([[0,1], [0,1], [0,1]])
    else:
        size_mults = np.asarray(size_mults)
    info += f'region box {region_style} {size_mults[0,0]} {size_mults[0,1]} {size_mults[1,0]} {size_mults[1,1]} {size_mults[2,0]} {size_mults[2,1]}'
    if region_style == 'prism':
        size_xy = vects[1][0] * (size_mults[0,1] - size_mults[0,0]) / system.box.a
        size_xz = vects[2][0] * (size_mults[0,1] - size_mults[0,0]) / system.box.a
        size_yz = vects[2][1] * (size_mults[1,1] - size_mults[1,0]) / system.box.b
        info += f' {size_xy} {size_xz} {size_yz}'
    info += '\n'
    
    # Create box and atoms
    info += f'create_box {natypes} box\n'
    info += f'create_atoms 1 box '
    for i in range(system.natoms):
        if system.atoms.atype[i] > 1:
            info += f' &\n             basis {i+1} {system.atoms.atype[i]}'
    info += '\n'
    
    # Set pair_info
    if return_pair_info is True:
        if potential is None:
            raise ValueError('return_pair_info = True requires that potential be given')
        
        info += '\n'
        info += potential.pair_info(symbols=system.symbols,
                                    masses=system.masses)
    
    return info


# coding: utf-8

# http://www.numpy.org/
import numpy as np

from .crystalsystem import ishexagonal

__all__ = ['plane3to4', 'plane4to3', 'vector3to4', 'vector4to3',
           'plane_crystal_to_cartesian',
           'vector_crystal_to_cartesian',
           'vector_primitive_to_conventional', 
           'vector_conventional_to_primitive']

def plane3to4(indices):
    """
    Converts 3-term Miller (hkl) plane indices to 4-term hexagonal (hkil)
    Miller-Bravias indices.
    
    Parameters
    ----------
    indices : np.ndarray
        (..., 3) array of Miller crystallographic indices.
   
    Returns
    -------
    np.ndarray
        (..., 4) array of Miller-Bravais crystallographic indices.
    
    Raises
    ------
    ValueError
        If indices dimensions are not (..., 3).
    """
    
    # Verify compatible indices
    indices = np.asarray(indices)
    if indices.shape[-1] != 3:
        raise ValueError('Invalid index dimensions')
    
    newindices = np.empty(indices.shape[:-1] + (4,))
    newindices[..., 0] = indices[..., 0]
    newindices[..., 1] = indices[..., 1]
    newindices[..., 2] = -(newindices[..., 0] + newindices[..., 1])
    newindices[..., 3] = indices[..., 2]
    
    return newindices

def plane4to3(indices):
    """
    Converts 4-term hexagonal Miller-Bravias (hkil) plane indices to 3-term 
    Miller (hkl) indices.
    
    Parameters
    ----------
    indices : np.ndarray of int
        (..., 4) array of Miller-Bravais crystallographic indices.
   
    Returns
    -------
    np.ndarray of int
        (..., 3) array of Miller crystallographic indices.
   
    Raises
    ------    
    ValueError
        If indices dimensions are not (..., 4), or if h+k+i != 0.
    """
    
    # Verify compatible indices
    indices = np.asarray(indices)
    if indices.shape[-1] != 4:
        raise ValueError('Invalid index dimensions')
    if not np.allclose(indices[...,:3].sum(axis=-1), 0.0):
        raise ValueError('Invalid indices: h+k+i != 0')
    
    newindices = np.empty(indices.shape[:-1] + (3,))
    newindices[..., 0] = indices[..., 0]
    newindices[..., 1] = indices[..., 1]
    newindices[..., 2] = indices[..., 3]
    
    return newindices

def vector3to4(indices):
    """
    Converts 3-term Miller [uvw] vector indices to 4-term hexagonal [uvtw]
    Miller-Bravias indices. 
    
    Parameters
    ----------
    indices : np.ndarray
        (..., 3) array of Miller crystallographic indices.
   
    Returns
    -------
    np.ndarray
        (..., 4) array of Miller-Bravais crystallographic indices.
    
    Raises
    ------
    ValueError
        If indices dimensions are not (..., 3).
    """
    # Verify compatible indices
    indices = np.asarray(indices)
    if indices.shape[-1] != 3:
        raise ValueError('Invalid index dimensions')
    
    # Transform 
    newindices = np.empty(indices.shape[:-1] + (4,))
    newindices[..., 0] = (2 * indices[..., 0] - indices[..., 1]) / 3
    newindices[..., 1] = (2 * indices[..., 1] - indices[..., 0]) / 3
    newindices[..., 2] = -(newindices[..., 0] + newindices[..., 1])
    newindices[..., 3] = indices[..., 2] 
    
    return newindices
    
def vector4to3(indices):
    """
    Converts 4-term hexagonal Miller-Bravias [uvtw] vector indices to 3-term 
    Miller [uvw] indices.
    
    Parameters
    ----------
    indices : np.ndarray of int
        (..., 4) array of Miller-Bravais crystallographic indices.
   
    Returns
    -------
    np.ndarray of int
        (..., 3) array of Miller crystallographic indices.
   
    Raises
    ------    
    ValueError
        If indices dimensions are not (..., 4), or if u+v+t != 0.
    """
    # Verify compatible indices
    indices = np.asarray(indices)
    if indices.shape[-1] != 4:
        raise ValueError('Invalid index dimensions')
    if not np.allclose(indices[...,:3].sum(axis=-1), 0.0):
        raise ValueError('Invalid indices: u+v+t != 0')
    
    # Transform
    newindices = np.empty(indices.shape[:-1] + (3,))
    newindices[..., 0] = 2 * indices[..., 0] + indices[..., 1]
    newindices[..., 1] = 2 * indices[..., 1] + indices[..., 0]
    newindices[..., 2] = indices[..., 3] 
    
    return newindices

def vector_crystal_to_cartesian(indices, box):
    """
    Converts crystal indices to Cartesian vectors relative
    to a given lattice box. 
    
    Parameters
    ----------
    indices : np.ndarray
        (..., 3) array of [uvw] Miller crystallographic indices or 
        (..., 4) array of [uvtw] Miller-Bravais crystallographic indices.
    box : atomman.Box
        Box that defines the lattice cell vectors to use. 
   
    Returns
    -------
    np.ndarray of float
        (..., 3) array of Cartesian vectors.
        
    Raises
    ------
    ValueError
        If indices dimensions are not (..., 3) or (..., 4), or if
        hexagonal indices given with non-hexagonal box.
    """
    indices = np.array(indices)
    
    # Convert 4-term Miller-Bravais to standard 3-term indices
    if indices.shape[-1] == 4: # pylint: disable=unsubscriptable-object
        if not ishexagonal(box):
            raise ValueError('Hexagonal indices given with non-hexagonal box')
        indices = vector4to3(indices)
    
    if indices.shape[-1] != 3:
        raise ValueError('Invalid index dimensions')

    return (np.outer(indices[...,0], box.vects[0]) 
          + np.outer(indices[...,1], box.vects[1]) 
          + np.outer(indices[...,2], box.vects[2])).reshape(indices.shape)
    
def vector_primitive_to_conventional(indices, setting='p'):
    """
    Converts crystal indices relative to a primitive cell 
    to indices relative to a conventional cell in a specified setting.
    
    Parameters
    ----------
    indices : np.ndarray
        (..., 3) array of [uvw] Miller crystallographic indices
        relative to the primitive cell
    setting : str
        Specifies the conventional cell setting: 'p' for primitive,
        'a', 'b', 'c' for side-centered, 'i' for body-centered, and
        'f' for face-centered.
   
    Returns
    -------
    np.ndarray of float
        (..., 3) array of [uvw] Miller crystallographic indices
        relative to the conventional cell
        
    Raises
    ------
    ValuenError
        If indices dimensions are not (..., 3) or if an unknown setting
        value is given.
    """

    # Verify compatible indices
    indices = np.asarray(indices)
    if indices.shape[-1] != 3:
        raise ValueError('Invalid index dimensions')

    lattice_vectors = {}
    lattice_vectors['p'] = np.array([[ 1.0, 0.0, 0.0],
                                     [ 0.0, 1.0, 0.0],
                                     [ 0.0, 0.0, 1.0]])
    lattice_vectors['a'] = np.array([[ 1.0, 0.0, 0.0],
                                     [ 0.0, 0.5, 0.5],
                                     [ 0.0,-0.5, 0.5]])
    lattice_vectors['b'] = np.array([[ 0.5, 0.0, 0.5],
                                     [ 0.0, 1.0, 0.0],
                                     [-0.5, 0.0, 0.5]])
    lattice_vectors['c'] = np.array([[ 0.5, 0.5, 0.0],
                                     [-0.5, 0.5, 0.0],
                                     [ 0.0, 0.0, 1.0]])
    lattice_vectors['i'] = np.array([[ 0.5, 0.5, 0.5],
                                     [-0.5, 0.5,-0.5],
                                     [-0.5,-0.5, 0.5]])
    lattice_vectors['f'] = np.array([[ 0.5, 0.5, 0.0],
                                     [ 0.5, 0.0, 0.5],
                                     [ 0.0, 0.5, 0.5]])
    
    try:
        lat = lattice_vectors[setting]
    except:
        raise ValueError('Unknown lattice setting. Allowed values are: p, a, b, c, i, and f')
    
    return np.inner(indices, lat.T)


def vector_conventional_to_primitive(indices, setting='p'):
    """
    Converts crystal indices relative to a conventional cell 
    in a specified setting to indices relative to a primitive cell.
    
    Parameters
    ----------
    indices : np.ndarray
        (..., 3) array of [uvw] Miller crystallographic indices
        relative to the conventional cell
    setting : str
        Specifies the conventional cell setting: 'p' for primitive,
        'a', 'b', 'c' for side-centered, 'i' for body-centered, and
        'f' for face-centered.
   
    Returns
    -------
    np.ndarray of float
        (..., 3) array of [uvw] Miller crystallographic indices
        relative to the primitive cell
        
    Raises
    ------
    ValuenError
        If indices dimensions are not (..., 3) or if an unknown setting
        value is given.
    """
    # Verify compatible indices
    indices = np.asarray(indices)
    if indices.shape[-1] != 3:
        raise ValueError('Invalid index dimensions')
    
    lattice_vectors = {}
    lattice_vectors['p'] = np.array([[ 1.0, 0.0, 0.0],
                                     [ 0.0, 1.0, 0.0],
                                     [ 0.0, 0.0, 1.0]])
    lattice_vectors['a'] = np.array([[ 1.0, 0.0, 0.0],
                                     [ 0.0, 1.0,-1.0],
                                     [ 0.0, 1.0, 1.0]])
    lattice_vectors['b'] = np.array([[ 1.0, 0.0,-1.0],
                                     [ 0.0, 1.0, 0.0],
                                     [ 1.0, 0.0, 1.0]])
    lattice_vectors['c'] = np.array([[ 1.0,-1.0, 0.0],
                                     [ 1.0, 1.0, 0.0],
                                     [ 0.0, 0.0, 1.0]])
    lattice_vectors['i'] = np.array([[ 0.0,-1.0,-1.0],
                                     [ 1.0, 1.0, 0.0],
                                     [ 1.0, 0.0, 1.0]])
    lattice_vectors['f'] = np.array([[ 1.0, 1.0,-1.0],
                                     [ 1.0,-1.0, 1.0],
                                     [-1.0, 1.0, 1.0]])
    
    try:
        lat = lattice_vectors[setting]
    except:
        raise ValueError('Unknown lattice setting. Allowed values are: p, a, b, c, i, and f')
    
    return np.inner(indices, lat.T)

def plane_crystal_to_cartesian(indices, box):
    """
    Converts crystal planar indices to Cartesian plane normal vectors relative
    to a given lattice box.  Note: the algorithm used requires that the planar
    indices be integers.
    
    Parameters
    ----------
    indices : np.ndarray
        (3,) array of [hkl] Miller crystallographic indices or 
        (4,) array of [hkil] Miller-Bravais crystallographic indices.
    box : atomman.Box
        Box that defines the lattice cell vectors to use. 
   
    Returns
    -------
    np.ndarray of float
        (..., 3) array of Cartesian vectors corresponding to plane normals.
        
    Raises
    ------
    ValueError
        If indices dimensions are not (..., 3) or (..., 4), or if
        hexagonal indices given with non-hexagonal box.
    """
    # Check indices values
    indices = np.asarray(indices)
    
    # Convert hkil to hkl
    if indices.shape == (4,):
        if ishexagonal(box):
            indices = plane4to3(indices)
        else:
            raise ValueError('Hexagonal indices given with non-hexagonal box')
    if indices.shape != (3,):
        raise ValueError('Invalid index dimensions')
    
    if np.allclose(indices, np.asarray(indices, dtype=int)):
        indices = np.asarray(indices, dtype=int)
    else:
        raise ValueError('Indices must be integers')
    
    # Find two in-plane box vectors
    if indices[0] != 0:
        if indices[1] != 0:
            if indices[2] != 0:
                # indices solution
                m = np.lcm.reduce([indices[0], indices[1], indices[2]]) # pylint: disable=no-member
                s = np.sign(indices[0] * indices[1] * indices[2])
                a_uvw = np.array([-m / indices[0], m / indices[1], 0], dtype=int)
                b_uvw = np.array([-m / indices[0], 0, m / indices[2]], dtype=int)
            else:
                # hk0 solution
                m = np.lcm(indices[0], indices[1]) # pylint: disable=no-member
                s = np.sign(indices[0] * indices[1])
                a_uvw = np.array([-m / indices[0], m / indices[1], 0], dtype=int)
                b_uvw = np.array([0, 0, 1], dtype=int)
        else:
            if indices[2] != 0:
                # h0l solution
                m = np.lcm(indices[0], indices[2]) # pylint: disable=no-member
                s = np.sign(indices[0] * indices[2])
                a_uvw = np.array([m / indices[0], 0, -m / indices[2]], dtype=int)
                b_uvw = np.array([0, 1, 0], dtype=int)
            else:
                # h00 solution
                m = 1
                s = np.sign(indices[0])
                a_uvw = np.array([0, 1, 0], dtype=int)
                b_uvw = np.array([0, 0, 1], dtype=int)
    elif indices[1] != 0:
        if indices[2] != 0:
            # 0kl solution
            m = np.lcm(indices[1], indices[2]) # pylint: disable=no-member
            s = np.sign(indices[1] * indices[2])
            a_uvw = np.array([0, -m / indices[1], m / indices[2]], dtype=int)
            b_uvw = np.array([1, 0, 0], dtype=int)
        else:    
            # 0k0 solution
            m = 1
            s = np.sign(indices[1])
            a_uvw = np.array([0, 0, 1], dtype=int)
            b_uvw = np.array([1, 0, 0], dtype=int)
    elif indices[2] != 0:
        # 00l solution
        m = 1
        s = np.sign(indices[2])
        a_uvw = np.array([1, 0, 0], dtype=int)
        b_uvw = np.array([0, 1, 0], dtype=int)
    else:
        raise ValueError('indices cannot be all zeros')
    
    # Compute Cartesian plane normal
    planenormal = s * np.cross(np.inner(a_uvw, box.vects.T),
                               np.inner(b_uvw, box.vects.T))

    # Return the unit vector normal
    return planenormal / np.linalg.norm(planenormal)